from aerosandbox.modeling.fitting import FittedModel
import pytest
import aerosandbox.numpy as np


def test_fit_model_weighting():
    x = np.linspace(0, 10)
    y = np.sin(x)

    fm = FittedModel(
        model=lambda x, p: p["m"] * x + p["b"],
        x_data=x,
        y_data=y,
        parameter_guesses={
            "m": 0,
            "b": 0,
        },
        weights=None
    )  # Fit a model with no weighting

    assert fm(10) != pytest.approx(5, abs=1)  # Doesn't give a high value at x = 10

    fm = FittedModel(
        model=lambda x, p: p["m"] * x + p["b"],
        x_data=x,
        y_data=y,
        parameter_guesses={
            "m": 0,
            "b": 0,
        },
        weights=(x > 0) & (x < 2)
    )  # Fit a model with weighting

    assert fm(10) == pytest.approx(5, abs=1)  # Gives a high value at x = 10

    fm.plot()


if __name__ == '__main__':
    test_fit_model_weighting()



import aerosandbox as asb
import pytest


def is_tool(name):
    """Check whether `name` is on PATH and marked as executable."""

    # from whichcraft import which
    from shutil import which

    return which(name) is not None


avl_present = is_tool('avl')


def test_conventional():
    if not avl_present:
        return
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.conventional import airplane
    analysis = asb.AVL(
        airplane=airplane,
        op_point=asb.OperatingPoint(alpha=10),
    )
    return analysis.run()


def test_vanilla():
    if not avl_present:
        return
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.vanilla import airplane
    analysis = asb.AVL(
        airplane=airplane,
        op_point=asb.OperatingPoint(alpha=10),
    )
    return analysis.run()


def test_flat_plate():
    if not avl_present:
        return
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.flat_plate import airplane
    analysis = asb.AVL(
        airplane=airplane,
        op_point=asb.OperatingPoint(alpha=10),
    )
    return analysis.run()


def test_flat_plate_mirrored():
    if not avl_present:
        return
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.flat_plate_mirrored import airplane
    analysis = asb.AVL(
        airplane=airplane,
        op_point=asb.OperatingPoint(alpha=10),
    )
    return analysis.run()


if __name__ == '__main__':
    # test_conventional()
    # test_vanilla()

    print(test_flat_plate()['CL'])
    # pytest.main()



import aerosandbox as asb
import aerosandbox.numpy as np
import aerosandbox.tools.units as u

# From Raymer, Aircraft Design: A Conceptual Approach, 5th Ed.
# Table 15.3: Miscellaneous Weights

mass_passenger = 215 * u.lbm  # includes carry-on


def mass_seat(
        kind="passenger"
) -> float:
    """
    Computes the mass of an individual seat on an airplane.

    Args:

        kind: The kind of seat. Can be "passenger", "flight_deck", or "troop".

            * "passenger" seats are standard commercial airline seats.

            * "flight_deck" seats are the seats in the cockpit.

            * "troop" seats are the seats in the cargo hold.

    Returns: The mass of a single seat, in kg. Don't forget to multiply by the number of seats to get the total mass
    of all seats.
    """
    if kind == "passenger":
        return 32 * u.lbm
    elif kind == "flight_deck":
        return 60 * u.lbm
    elif kind == "troop":
        return 11 * u.lbm
    else:
        raise ValueError("Bad value of `kind`!")


def mass_lavatories(
        n_pax,
        aircraft_type="short-haul"
) -> float:
    """
    Computes the required mass of all lavatories on an airplane.

    Args:
        n_pax: The number of passengers on the airplane.

        aircraft_type: The type of aircraft. Can be "long-haul", "short-haul", or "business-jet".

            * "long-haul" aircraft are long-range commercial airliners, like the Boeing 777 or Airbus A350.

            * "short-haul" aircraft are short-range commercial airliners, like the Boeing 737 or Airbus A320.

            * "business-jet" aircraft are small private jets, like the Cessna Citation X or Gulfstream G650.

    Returns: The mass of all lavatories on the airplane, in kg.

    """
    if aircraft_type == "long-haul":
        return (1.11 * n_pax ** 1.33) * u.lbm
    elif aircraft_type == "short-haul":
        return (0.31 * n_pax ** 1.33) * u.lbm
    elif aircraft_type == "business-jet":
        return (3.90 * n_pax ** 1.33) * u.lbm
    else:
        raise ValueError("Bad value of `aircraft_type`!")




from typing import Union, Dict, Any
import aerosandbox.numpy as np
from aerosandbox.modeling.interpolation import InterpolatedModel
from scipy import interpolate


class UnstructuredInterpolatedModel(InterpolatedModel):
    """
    A model that is interpolated to unstructured (i.e., point cloud) N-dimensional data. Maps from R^N -> R^1.

    You can evaluate this model at a given point by calling it just like a function, e.g.:

    >>> y = my_interpolated_model(x)

    The input to the model (`x` in the example above) is of the type:
        * in the general N-dimensional case, a dictionary where: keys are variable names and values are float/array
        * in the case of a 1-dimensional input (R^1 -> R^1), it can optionally just be a float/array.
    If you're not sure what the input type of `my_interpolated_model` should be, just do:

    >>> print(my_interpolated_model) # Displays the valid input type to the model

    The output of the model (`y` in the example above) is always a float or array.

    See the docstring __init__ method of InterpolatedModel for more details of how to instantiate and use UnstructuredInterpolatedModel.

    """

    def __init__(self,
                 x_data: Union[np.ndarray, Dict[str, np.ndarray]],
                 y_data: np.ndarray,
                 x_data_resample: Union[int, Dict[str, Union[int, np.ndarray]]] = 10,
                 resampling_interpolator: object = interpolate.RBFInterpolator,
                 resampling_interpolator_kwargs: Dict[str, Any] = None,
                 fill_value=np.nan,  # Default behavior: return NaN for all inputs outside data range.
                 interpolated_model_kwargs: Dict[str, Any] = None,
                 ):
        """
        Creates the interpolator. Note that data must be unstructured (i.e., point cloud) for general N-dimensional
        interpolation.

        Note that if data is either 1D or structured,

        Args:

            x_data: Values of the dependent variable(s) in the dataset to be fitted. This is a dictionary; syntax is {
            var_name:var_data}.

                * If the model is one-dimensional (e.g. f(x1) instead of f(x1, x2, x3...)), you can instead supply x_data
                as a 1D ndarray. (If you do this, just treat `x` as an array in your model, not a dict.)

            y_data: Values of the independent variable in the dataset to be fitted. [1D ndarray of length n]

            x_data_resample: A parameter that guides how the x_data should be resampled onto a structured grid.

                * If this is an int, we look at each axis of the `x_data` (here, we'll call this `xi`),
                and we resample onto a linearly-spaced grid between `min(xi)` and `max(xi)` with `x_data_resample`
                points.

                * If this is a dict, it must be a dict where the keys are strings matching the keys of (the
                dictionary) `x_data`. The values can either be ints or 1D np.ndarrays.

                    * If the values are ints, then that axis is linearly spaced between `min(xi)` and `max(xi)` with
                    `x_data_resample` points.

                    * If the values are 1D np.ndarrays, then those 1D np.ndarrays are used as the resampled spacing
                    for the given axis.

            resampling_interpolator: Indicates the interpolator to use in order to resample the unstructured data
            onto a structured grid. Should be analogous to scipy.interpolate.RBFInterpolator in __init__ and __call__
            syntax. See reference here:

                * https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.RBFInterpolator.html

            resampling_interpolator_kwargs: Indicates keyword arguments (keyword-value pairs, as a dictionary) to
            pass into the resampling interpolator.

            fill_value: Gives the value that the interpolator should return for points outside of the interpolation
            domain. The interpolation domain is defined as the hypercube bounded by the coordinates specified in
            `x_data_resample`. By default, these coordinates are the tightest axis-aligned hypercube that bounds the
            point cloud data. If fill_value is None, then the interpolator will attempt to extrapolate if the interpolation method allows.

            interpolated_model_kwargs: Indicates keyword arguments to pass into the (structured) InterpolatedModel.
            Also a dictionary. See aerosandbox.InterpolatedModel for documentation on possible inputs here.

        """
        if resampling_interpolator_kwargs is None:
            resampling_interpolator_kwargs = {}
        if interpolated_model_kwargs is None:
            interpolated_model_kwargs = {}

        try:  # Try to use the InterpolatedModel initializer. If it doesn't work, then move on.
            super().__init__(
                x_data_coordinates=x_data,
                y_data_structured=y_data,
            )
            return
        except ValueError:
            pass

        # If it didn't work, this implies that x_data is multidimensional, and hence a dict-like object. Validate this.
        try:  # Determine type of `x_data`
            x_data.keys()
            x_data.values()
            x_data.items()
        except AttributeError:
            raise TypeError("`x_data` must be a dict-like object!")

        # Make the interpolator, based on x_data and y_data.
        if resampling_interpolator == interpolate.RBFInterpolator:
            resampling_interpolator_kwargs = {
                "kernel": "thin_plate_spline",
                "degree": 1,
                **resampling_interpolator_kwargs
            }

        interpolator = resampling_interpolator(
            y=np.stack(tuple(x_data.values()), axis=1),
            d=y_data,
            **resampling_interpolator_kwargs
        )

        # If x_data_resample is an int, make it into a dict that matches x_data.
        if isinstance(x_data_resample, int):
            x_data_resample = {
                k: x_data_resample
                for k in x_data.keys()
            }

        # Now, x_data_resample should be dict-like. Validate this.
        try:
            x_data_resample.keys()
            x_data_resample.values()
            x_data_resample.items()
        except AttributeError:
            raise TypeError("`x_data_resample` must be a dict-like object!")

        # Go through x_data_resample, and replace any values that are ints with linspaced arrays.
        for k, v in x_data_resample.items():
            if isinstance(v, int):
                x_data_resample[k] = np.linspace(
                    np.min(x_data[k]),
                    np.max(x_data[k]),
                    v
                )

        x_data_coordinates: Dict = x_data_resample

        x_data_structured_values = [
            xi.flatten()
            for xi in np.meshgrid(*x_data_coordinates.values(), indexing="ij")
        ]
        x_data_structured = {
            k: xi
            for k, xi in zip(x_data.keys(), x_data_structured_values)
        }

        y_data_structured = interpolator(
            np.stack(tuple(x_data_structured_values), axis=1)
        )
        y_data_structured = y_data_structured.reshape([
            np.length(xi)
            for xi in x_data_coordinates.values()
        ])

        interpolated_model_kwargs = {
            "fill_value": fill_value,
            **interpolated_model_kwargs
        }

        super().__init__(
            x_data_coordinates=x_data_coordinates,
            y_data_structured=y_data_structured,
            **interpolated_model_kwargs,
        )

        self.x_data_raw_unstructured = x_data
        self.y_data_raw = y_data


if __name__ == '__main__':
    x = np.arange(10)
    y = x ** 3
    interp = UnstructuredInterpolatedModel(
        x_data=x,
        y_data=y
    )


    def randspace(start, stop, n=50):
        vals = (stop - start) * np.random.rand(n) + start
        vals = np.concatenate((vals[:-2], np.array([start, stop])))
        # vals = np.sort(vals)
        return vals


    np.random.seed(4)
    X = randspace(-5, 5, 200)
    Y = randspace(-5, 5, 200)
    f = np.where(X > 0, 1, 0) + np.where(Y > 0, 1, 0)
    # f = X ** 2 + Y ** 2
    interp = UnstructuredInterpolatedModel(
        x_data={
            "x": X.flatten(),
            "y": Y.flatten(),
        },
        y_data=f.flatten()
    )

    from aerosandbox.tools.pretty_plots import plt, show_plot

    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    # ax.plot_surface(X, Y, f, color="blue", alpha=0.2)
    ax.scatter(X.flatten(), Y.flatten(), f.flatten())
    X_plot, Y_plot = np.meshgrid(
        np.linspace(X.min(), X.max(), 500),
        np.linspace(Y.min(), Y.max(), 500),
    )
    F_plot = interp({
        "x": X_plot.flatten(),
        "y": Y_plot.flatten()
    }).reshape(X_plot.shape)
    ax.plot_surface(
        X_plot, Y_plot, F_plot,
        color="red",
        edgecolors=(1, 1, 1, 0.5),
        linewidth=0.5,
        alpha=0.2,
        rcount=40,
        ccount=40,
        shade=True,
    )
    plt.xlabel("x")
    plt.ylabel("y")
    plt.show()

    import aerosandbox as asb
    import aerosandbox.numpy as np

    opti = asb.Opti()
    x = opti.variable(init_guess=0)
    y = opti.variable(init_guess=0)
    opti.minimize(interp({"x": x, "y": y}))
    sol = opti.solve()
    print(sol(x))
    print(sol(y))



import aerosandbox as asb
import aerosandbox.numpy as np
import cadquery as cq
import copy
from typing import List, Union, Dict
from sortedcontainers import SortedDict


class WingStructureGenerator():
    def __init__(self,
                 wing: asb.Wing,
                 default_rib_thickness=3e-3,
                 minimum_airfoil_TE_thickness_rel: float = 0.001
                 ):
        self.wing = wing
        self.default_rib_thickness = default_rib_thickness
        self.minimum_airfoil_TE_thickness_rel = minimum_airfoil_TE_thickness_rel

        ### Compute some span properties which are used for locating ribs
        self._sectional_spans: List[float] = wing.span(_sectional=True)
        self._cumulative_spans_up_to_section = np.concatenate((
            [0],
            np.cumsum(self._sectional_spans)
        ))
        self._total_span = sum(self._sectional_spans)

        ### Generate the OML geometry
        self.oml = asb.Airplane(
            wings=[wing]
        ).generate_cadquery_geometry(
            minimum_airfoil_TE_thickness=minimum_airfoil_TE_thickness_rel
        )

        ### Set up data structures for geometry
        self.ribs: SortedDict[cq.Workplane] = SortedDict()
        self.spars: List[cq.Workplane] = []

    def __repr__(self):
        return f"{self.__class__.__name__}({self.wing})"

    # def open_interactive(self):

    def add_ribs_from_section_span_fractions(
            self,
            section_index: int,
            section_span_fractions: Union[float, int, List[float], np.ndarray],
            rib_thickness: float = None,
    ):
        if rib_thickness is None:
            rib_thickness = self.default_rib_thickness

        try:
            iter(section_span_fractions)
        except TypeError:
            section_span_fractions = [section_span_fractions]

        xsec_a = wing.xsecs[section_index]
        xsec_b = wing.xsecs[section_index + 1]

        for s in section_span_fractions:
            af = xsec_a.airfoil.blend_with_another_airfoil(
                airfoil=xsec_b.airfoil,
                blend_fraction=s
            )

            chord = (
                    (1 - s) * xsec_a.chord + s * xsec_b.chord
            )

            csys = wing._compute_frame_of_section(section_index)

            span = (
                    self._cumulative_spans_up_to_section[section_index]
                    + s * self._sectional_spans[section_index]
            )

            self.ribs[span] = (
                cq.Workplane(
                    inPlane=cq.Plane(
                        origin=tuple(
                            (1 - s) * xsec_a.xyz_le + s * xsec_b.xyz_le
                        ),
                        xDir=tuple(csys[0]),
                        normal=tuple(-csys[1])
                    )
                ).spline(
                    listOfXYTuple=[
                        tuple(xy * chord)
                        for xy in af.coordinates
                    ]
                ).close().extrude(
                    rib_thickness / 2,
                    combine=False,
                    both=True
                )
            )

    def add_ribs_from_xsecs(
            self,
            indexes: List[int] = None,
            rib_thickness: float = None
    ):
        if rib_thickness is None:
            rib_thickness = self.default_rib_thickness

        if indexes is None:
            indexes = range(len(wing.xsecs))

        for i in indexes:
            xsec = wing.xsecs[i]
            csys = wing._compute_frame_of_WingXSec(i)

            af = xsec.airfoil
            if af.TE_thickness() < self.minimum_airfoil_TE_thickness_rel:
                af = af.set_TE_thickness(
                    thickness=self.minimum_airfoil_TE_thickness_rel
                )

            span = self._cumulative_spans_up_to_section[i]

            self.ribs[span] = (
                cq.Workplane(
                    inPlane=cq.Plane(
                        origin=tuple(xsec.xyz_le),
                        xDir=tuple(csys[0]),
                        normal=tuple(-csys[1])
                    )
                ).spline(
                    listOfXYTuple=[
                        tuple(xy * xsec.chord)
                        for xy in af.coordinates
                    ]
                ).close().extrude(
                    rib_thickness / 2,
                    combine=False,
                    both=True
                )
            )

    def add_ribs_from_span_fractions(
            self,
            span_fractions: Union[float, List[float], np.ndarray] = np.linspace(0, 1, 10),
            rib_thickness: float = None,
    ):
        ### Handle span_fractions if it's not an iterable
        try:
            iter(span_fractions)
        except TypeError:
            span_fractions = [span_fractions]

        for s in span_fractions:
            if s == 0:
                section_index = 0
                section_span_fraction = 0
            elif s == 1:
                section_index = len(self.wing.xsecs) - 2
                section_span_fraction = 1
            elif s < 0 or s > 1:
                raise ValueError(
                    "All values of `span_fractions` must be between 0 and 1!"
                )
            else:
                section_index = np.argwhere(
                    self._cumulative_spans_up_to_section > self._total_span * s
                )[0][0] - 1

                section_span_fraction = (
                                                s * self._total_span
                                                - self._cumulative_spans_up_to_section[section_index]
                                        ) / self._sectional_spans[section_index]

            self.add_ribs_from_section_span_fractions(
                section_index=section_index,
                section_span_fractions=section_span_fraction,
                rib_thickness=rib_thickness
            )

    def add_tube_spar(self,
                      span_location_root: float,
                      span_location_tip: float,
                      diameter_root,
                      x_over_c_location_root=0.25,
                      y_over_c_location_root=None,
                      x_over_c_location_tip=None,
                      y_over_c_location_tip=None,
                      diameter_tip: float = None,
                      cut_ribs: bool = True,
                      ):
        if diameter_tip is None:
            diameter_tip = diameter_root
        if x_over_c_location_tip is None:
            x_over_c_location_tip = x_over_c_location_root
            # TODO change behavior so that default is a 90 degree spar
        if y_over_c_location_root is None:
            y_over_c_location_root = wing.xsecs[0].airfoil.local_camber(
                x_over_c=x_over_c_location_root
            )
        if y_over_c_location_tip is None:
            y_over_c_location_tip = y_over_c_location_root
            # TODO change behavior so that default is a 90 degree spar

        ### Figure out where the spar root is
        section_index = np.argwhere(
            self._cumulative_spans_up_to_section > span_location_root
        )[0][0] - 1

        section_span_fraction = (
                                        span_location_root
                                        - self._cumulative_spans_up_to_section[section_index]
                                ) / self._sectional_spans[section_index]

        root_csys = self.wing._compute_frame_of_section(section_index)

        root_le_point = (
                (1 - section_span_fraction) * wing.xsecs[section_index].xyz_le
                + section_span_fraction * wing.xsecs[section_index + 1].xyz_le
        )
        root_chord = (
                (1 - section_span_fraction) * wing.xsecs[section_index].chord
                + section_span_fraction * wing.xsecs[section_index + 1].chord
        )
        root_point = (
                root_le_point +
                x_over_c_location_root * root_csys[0] * root_chord +
                y_over_c_location_root * root_csys[2] * root_chord
        )

        ### Figure out where the spar tip is

        section_index = np.argwhere(
            self._cumulative_spans_up_to_section > span_location_tip
        )[0][0] - 1

        section_span_fraction = (
                                        span_location_tip
                                        - self._cumulative_spans_up_to_section[section_index]
                                ) / self._sectional_spans[section_index]

        tip_csys = self.wing._compute_frame_of_section(section_index)

        tip_le_point = (
                (1 - section_span_fraction) * wing.xsecs[section_index].xyz_le
                + section_span_fraction * wing.xsecs[section_index + 1].xyz_le
        )
        tip_chord = (
                (1 - section_span_fraction) * wing.xsecs[section_index].chord
                + section_span_fraction * wing.xsecs[section_index + 1].chord
        )
        tip_point = (
                tip_le_point +
                x_over_c_location_tip * tip_csys[0] * tip_chord +
                y_over_c_location_tip * tip_csys[2] * tip_chord
        )

        normal = tip_point - root_point

        root_plane = cq.Plane(
            origin=tuple(root_point),
            xDir=tuple(root_csys[0]),
            normal=tuple(normal)
        )

        tip_plane = cq.Plane(
            origin=tuple(tip_point),
            xDir=tuple(tip_csys[0]),
            normal=tuple(normal)
        )

        ### Make the spar
        root_wire = cq.Workplane(
            inPlane=root_plane
        ).circle(radius=2 * diameter_root / 2)
        tip_wire = cq.Workplane(
            inPlane=tip_plane
        ).circle(radius=2 * diameter_tip / 2)

        wire_collection = root_wire
        wire_collection.ctx.pendingWires.extend(tip_wire.ctx.pendingWires)

        spar = wire_collection.loft(ruled=True, clean=False)

        self.spars.append(spar)

        if cut_ribs:
            for span_loc, rib in self.ribs.items():
                rib: cq.Workplane
                self.ribs[span_loc] = rib.cut(spar)


from aerosandbox.tools import units as u

af = asb.Airfoil("sd7032")

wing = asb.Wing(
    name="Wing",
    symmetric=True,
    xsecs=[
        asb.WingXSec(
            chord=0.2,
            airfoil=af
        ),
        asb.WingXSec(
            xyz_le=[0.05, 0.5, 0],
            chord=0.15,
            airfoil=af
        ),
        asb.WingXSec(
            xyz_le=[0.1, 0.8, 0.1],
            chord=0.1,
            airfoil=af
        )
    ]
)

s = WingStructureGenerator(
    wing,
    default_rib_thickness=1 / 16 * u.inch
)
# s.add_ribs_from_section_span_fractions(0, np.linspace(0, 1, 10)[1:-1])
# s.add_ribs_from_section_span_fractions(1, np.linspace(0, 1, 8)[1:-1])
s.add_ribs_from_xsecs()
# s.add_ribs_from_span_fractions(
#     span_fractions=np.linspace(0, 1, 6)
# )

s.add_tube_spar(
    diameter_root=8e-3,
    span_location_root=-1 / 16 * u.inch,
    span_location_tip=s.ribs.keys()[1],
    diameter_tip=4e-3
)

### Show all ribs
rs = s.ribs.values()[0]
for rib in s.ribs.values()[1:]:
    rs = rs.union(rib)

ss = s.spars[0]

# oml = s.oml



import aerosandbox.numpy as np
from aerosandbox.geometry.fuselage import Fuselage, FuselageXSec
from textwrap import indent, dedent
from aerosandbox.geometry.openvsp_io.asb_to_openvsp import _utilities


def generate_fuselage(
        fuselage: Fuselage,
        include_main=True,
        continuity_type: str = "C2",
) -> str:
    """
    Generates a VSPScript file for a Fuselage object.

    Args:
        fuselage: The Fuselage object to generate a VSPScript file for.

    Returns: A VSPScript file as a string.

    """
    script = ""

    script += f"""\
//==== Add Fuselage "{fuselage.name}" ====//
string fid = AddGeom("FUSELAGE");
ChangeXSecShape( GetXSecSurf( fid, 0 ), 0, XS_SUPER_ELLIPSE );
ChangeXSecShape( GetXSecSurf( fid, 4 ), 4, XS_SUPER_ELLIPSE );

"""

    script += """//==== Generate Blank Fuselage Sections ====//\n"""
    if len(fuselage.xsecs) < 2:
        raise ValueError("Fuselages must have at least 2 cross sections.")
    else:
        script += f"""\
SetParmVal( fid, "XLocPercent", "XSec_3", 1.0 ); // Causes all sections to be generated at x/L = 1, so they don't conflict
        """
        for i in range(len(fuselage.xsecs) - 1):
            script += f"""\
InsertXSec( fid, 3, XS_SUPER_ELLIPSE ); // FuselageXSecs {i} to {i + 1}
"""

    length = fuselage.xsecs[-1].xyz_c[0] - fuselage.xsecs[0].xyz_c[0]
    script += f"""\
//===== Cut The Original Section =====//
CutXSec( fid, 0 );
CutXSec( fid, 0 );
CutXSec( fid, 0 );
CutXSec( fid, 0 );

//==== Set Overall Fuselage Options and First Section ====//
SetParmVal( fid, "X_Rel_Location", "XForm", {fuselage.xsecs[0].xyz_c[0]} );
SetParmVal( fid, "Y_Rel_Location", "XForm", {fuselage.xsecs[0].xyz_c[1]} );
SetParmVal( fid, "Z_Rel_Location", "XForm", {fuselage.xsecs[0].xyz_c[2]} );
SetParmVal( fid, "Length", "Design", {length} );
Update();

"""
    # TODO symmetry here

    script += """//==== Set Fuselage Section Options ====//\n"""
    if continuity_type == "C0":
        continuity_type_string = "0.0"
    elif continuity_type == "C1":
        continuity_type_string = "1.0"
    elif continuity_type == "C2":
        continuity_type_string = "2.0"
    else:
        raise ValueError("Continuity type must be 'C0', 'C1', or 'C2'.")

    for i, xsec in enumerate(fuselage.xsecs):
        script += f"""\
// ASB Section {i}, VSP Section {i}
SetParmVal( fid, "XLocPercent", "XSec_{i}", {xsec.xyz_c[0] / length} );
SetParmVal( fid, "YLocPercent", "XSec_{i}", {xsec.xyz_c[1] / length} );
SetParmVal( fid, "ZLocPercent", "XSec_{i}", {xsec.xyz_c[2] / length} );
SetParmVal( fid, "SectTess_U", "XSec_{i}", 10 );
SetParmVal( fid, "AllSym", "XSec_{i}", 1.0 );
SetParmVal( fid, "TopLAngleSet", "XSec_{i}", 0.0 );
SetParmVal( fid, "TopLAngle", "XSec_{i}", 0.0 );
SetParmVal( fid, "TopLStrengthSet", "XSec_{i}", 0.0 );
SetParmVal( fid, "TopLStrength", "XSec_{i}", 0.5 );
SetParmVal( fid, "ContinuityTop", "XSec_{i}", {continuity_type_string} );
Update();

"""

    script += """\
//==== Set Shapes ====//
"""

    for i in range(len(fuselage.xsecs)):
        xsec = fuselage.xsecs[i]
        script += f"""\
{{
    string xsec_surf = GetXSecSurf( fid, {i} );
    xsec_surf = GetXSecSurf( fid, {i} );
    string xsec = GetXSec( xsec_surf, {i} );
    SetParmVal( fid, "Super_Width", "XSecCurve_{i}", {xsec.width} );
    SetParmVal( fid, "Super_Height", "XSecCurve_{i}", {xsec.height} );
    SetParmVal( fid, "Super_M", "XSecCurve_{i}", {xsec.shape} );
    SetParmVal( fid, "Super_N", "XSecCurve_{i}", {xsec.shape} );
    Update();
}}
"""

    script += """\
SetParmVal( fid, "Tess_W", "Shape", 100 );
Update();
    """

    if include_main:
        script = _utilities.wrap_script(script)

    return script


if __name__ == '__main__':
    import aerosandbox as asb
    import aerosandbox.numpy as np

    af = asb.Airfoil("ls013")
    x = np.linspace(0, 1, 10)

    fuse = Fuselage(
        name="Fuse",
        xsecs=[
            FuselageXSec(
                xyz_c=[xi, 0, 0.05 * xi ** 2],
                width=2 * af.local_thickness(xi),
                height=af.local_thickness(xi),
                shape=4
            )
            for xi in x
        ]
    )
    print(generate_fuselage(fuse))

    with open("test.vspscript", "w") as f:
        f.write(generate_fuselage(fuse))



import numpy as _onp
import casadi as _cas
from typing import Tuple, Iterable, Union
from aerosandbox.numpy.conditionals import where

from aerosandbox.numpy.determine_type import is_casadi_type


def _make_casadi_types_broadcastable(x1, x2):
    def shape_2D(object: Union[float, int, Iterable, _onp.ndarray]) -> Tuple:
        shape = _onp.shape(object)
        if len(shape) == 0:
            return (1, 1)
        elif len(shape) == 1:
            return (1, shape[0])
        elif len(shape) == 2:
            return shape
        else:
            raise ValueError("CasADi can't handle arrays with >2 dimensions, unfortunately.")

    x1_shape = shape_2D(x1)
    x2_shape = shape_2D(x2)
    shape = _onp.broadcast_shapes(x1_shape, x2_shape)

    x1_tiled = _cas.repmat(
        x1,
        shape[0] // x1_shape[0],
        shape[1] // x1_shape[1],
    )
    x2_tiled = _cas.repmat(
        x2,
        shape[0] // x2_shape[0],
        shape[1] // x2_shape[1],
    )

    return x1_tiled, x2_tiled


def add(
        x1, x2
):
    if not is_casadi_type(x1) and not is_casadi_type(x2):
        return _onp.add(x1, x2)
    else:
        x1, x2 = _make_casadi_types_broadcastable(x1, x2)
        return x1 + x2


def multiply(
        x1, x2
):
    if not is_casadi_type(x1) and not is_casadi_type(x2):
        return _onp.multiply(x1, x2)
    else:
        x1, x2 = _make_casadi_types_broadcastable(x1, x2)
        return x1 * x2


def mod(x1, x2):
    """
    Return element-wise remainder of division.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.mod.html
    """
    if not is_casadi_type(x1) and not is_casadi_type(x2):
        return _onp.mod(x1, x2)

    else:
        out = _cas.fmod(x1, x2)
        out = where(
            x1 < 0,
            out + x2,
            out
        )
        return out


def centered_mod(x1, x2):
    """
    Return element-wise remainder of division, centered on zero.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.mod.html
    """
    if not is_casadi_type(x1) and not is_casadi_type(x2):
        remainder = _onp.mod(x1, x2)
        return where(
            remainder > x2 / 2,
            remainder - x2,
            remainder
        )

    else:
        return _cas.remainder(x1, x2)



from typing import Union
import casadi as _cas
import numpy as _onp
from aerosandbox.numpy.array import length, roll, concatenate
from aerosandbox.numpy.calculus import diff


def integrate_discrete_intervals(
        f: Union[_onp.ndarray, _cas.MX],
        x: Union[_onp.ndarray, _cas.MX] = None,
        multiply_by_dx: bool = True,
        method: str = "trapezoidal",
        method_endpoints: str = "lower_order",
):
    """
    Given a set of sampled points (x_i, f_i) from a function, computes the integral of that function over each set of
    adjacent points ("intervals"). Does this via a reconstruction approach, with several methods available.

    In general, N points will yield N-1 integrals (one for each "interval" between points).

    Args:
        f: A 1D array of function values.

        x: A 1D array of x-values where the function was evaluated. If not specified, defaults to the indices of f.
            Should be the same length as f and should be monotonically increasing (i.e. x[i] < x[i+1], with no duplicated points).

        multiply_by_dx: Whether to multiply the integral by the width of the segment. Defaults to True.
            - If True, summing the integrals will yield the integral of the function over the entire domain (x[0] to x[-1])
            - If False, you can think of the output as the "average function value" over each interval.

        method: The integration method to use. Options are:
            - "forward_euler"
            - "backward_euler"
            - "trapezoidal" (default)
            - "forward_simpson"
            - "backward_simpson"
            - "cubic"
            Note that some methods, like "cubic", approximate each segment interval by looking beyond just the integral itself (i.e., f(a) and f(b)),
             and so are not possible near the endpoints of the array.

        method_endpoints: The integration method to use at the endpoints, for those higher-order methods that require handling. Options are:
            - "lower_order" (default)
            - "ignore" (i.e. return the integral of the interior points only - note that this may result in a different number of integrals than segments!)
            - "periodic"

    """
    # Determine if an x-array was specified, and calculate dx.
    x_is_specified = (x is not None)
    if not x_is_specified:
        x = _onp.arange(length(f))

    dx = x[1:] - x[:-1]

    method = str(method).lower().replace(" ", "_")

    # Implement integration methods
    if method in ["forward_euler", "forward", "euler_forward", "left", "left_riemann"]:
        avg_f = f[:-1]

        degree = 0  # Refers to the highest degree of the polynomial that the method is exact for.
        remaining_endpoint_intervals = (0, 0)

    elif method in ["backward_euler", "backward", "euler_backward", "right", "right_riemann"]:
        avg_f = f[1:]

        degree = 0
        remaining_endpoint_intervals = (0, 0)

    elif method in ["trapezoidal", "trapezoid", "trapz", "midpoint"]:
        if method == "midpoint":
            raise PendingDeprecationWarning(
                "The 'midpoint' method will be deprecated at a future point, since 'trapezoidal' is the more accurate term here.")

        avg_f = (f[1:] + f[:-1]) / 2

        degree = 1
        remaining_endpoint_intervals = (0, 0)

    elif method in ["forward_simpson", "simpson_forward", "simpson"]:
        x1 = x[:-2]
        x2 = x[1:-1]
        x3 = x[2:]

        f1 = f[:-2]
        f2 = f[1:-1]
        f3 = f[2:]

        h = x2 - x1
        hp = x3 - x2

        # q1 = 0  # Integration lower bound
        # q2 = 1  # Integration upper bound
        q3 = 1 + hp / h

        avg_f = (
                        f1 - f3
                        + 3 * q3 ** 2 * (f1 + f2)
                        - 2 * q3 * (2 * f1 + f2)
                ) / (
                        6 * q3 * (q3 - 1)
                )

        degree = 2
        remaining_endpoint_intervals = (0, 1)

    elif method in ["backward_simpson", "simpson_backward"]:
        x1 = x[:-2]
        x2 = x[1:-1]
        x3 = x[2:]

        f1 = f[:-2]
        f2 = f[1:-1]
        f3 = f[2:]

        h = x3 - x2
        hm = x2 - x1

        q1 = -hm / h
        # q2 = 0  # Integration lower bound
        # q3 = 1  # Integration upper bound

        avg_f = (
                        f2 - f1
                        + 3 * q1 ** 2 * (f2 + f3)
                        - 2 * q1 * (2 * f2 + f3)
                ) / (
                        6 * q1 * (q1 - 1)
                )

        degree = 2
        remaining_endpoint_intervals = (1, 0)

    elif method in ["cubic", "cubic_spline"]:
        x1 = x[:-3]
        x2 = x[1:-2]
        x3 = x[2:-1]
        x4 = x[3:]

        f1 = f[:-3]
        f2 = f[1:-2]
        f3 = f[2:-1]
        f4 = f[3:]

        h = x3 - x2
        hm = x2 - x1
        hp = x4 - x3

        q1 = -hm / h
        # q2 = 0  # Integration lower bound
        # q3 = 1  # Integration upper bound
        q4 = 1 + hp / h

        avg_f = (
                        6 * q1 ** 3 * q4 ** 2 * (f2 + f3)
                        - 4 * q1 ** 3 * q4 * (2 * f2 + f3)
                        + 2 * q1 ** 3 * (f2 - f4)
                        - 6 * q1 ** 2 * q4 ** 3 * (f2 + f3)
                        + 3 * q1 ** 2 * q4 * (3 * f2 + f3)
                        + 3 * q1 ** 2 * (f4 - f2)
                        + 4 * q1 * q4 ** 3 * (2 * f2 + f3)
                        - 3 * q1 * q4 ** 2 * (3 * f2 + f3)
                        + q1 * (f2 - f4)
                        + 2 * q4 ** 3 * (f1 - f2)
                        + 3 * q4 ** 2 * (f2 - f1)
                        + q4 * (f1 - f2)
                ) / (
                        12 * q1 * q4 * (q1 - 1) * (q1 - q4) * (q4 - 1)
                )

        degree = 3
        remaining_endpoint_intervals = (1, 1)

    else:
        raise ValueError(f"Invalid method '{method}'.")

    if method_endpoints == "lower_order":
        if degree >= 3:
            method_endpoints = "simpson"
        else:
            method_endpoints = "trapezoidal"

        if method_endpoints == "simpson":

            if remaining_endpoint_intervals[0] != 0:
                avg_f_left_intervals = integrate_discrete_intervals(
                    f=f[:2 + remaining_endpoint_intervals[0]],
                    x=x[:2 + remaining_endpoint_intervals[0]],
                    multiply_by_dx=False,
                    method="forward_simpson",
                    method_endpoints="ignore",
                )
                avg_f = concatenate((
                    avg_f_left_intervals,
                    avg_f
                ))

            if remaining_endpoint_intervals[1] != 0:
                avg_f_right_intervals = integrate_discrete_intervals(
                    f=f[-(2 + remaining_endpoint_intervals[1]):],
                    x=x[-(2 + remaining_endpoint_intervals[1]):],
                    multiply_by_dx=False,
                    method="backward_simpson",
                    method_endpoints="ignore",
                )
                avg_f = concatenate((
                    avg_f,
                    avg_f_right_intervals,
                ))

        elif method_endpoints == "trapezoidal":

            if remaining_endpoint_intervals[0] != 0:
                avg_f_left_intervals = integrate_discrete_intervals(
                    f=f[:1 + remaining_endpoint_intervals[0]],
                    x=x[:1 + remaining_endpoint_intervals[0]],
                    multiply_by_dx=False,
                    method="trapezoidal",
                    method_endpoints="ignore",
                )

                avg_f = concatenate((
                    avg_f_left_intervals,
                    avg_f
                ))

            if remaining_endpoint_intervals[1] != 0:
                avg_f_right_intervals = integrate_discrete_intervals(
                    f=f[-(1 + remaining_endpoint_intervals[1]):],
                    x=x[-(1 + remaining_endpoint_intervals[1]):],
                    multiply_by_dx=False,
                    method="trapezoidal",
                    method_endpoints="ignore",
                )

                avg_f = concatenate((
                    avg_f,
                    avg_f_right_intervals,
                ))

        else:
            raise ValueError(f"Invalid method_endpoints '{method_endpoints}'.")

    elif method_endpoints == "ignore":
        pass

    elif method_endpoints == "periodic":
        raise NotImplementedError("Periodic integration is not yet implemented.")

    else:
        raise ValueError(f"Invalid method_endpoints '{method_endpoints}'.")

    if multiply_by_dx:
        if x_is_specified:
            return avg_f * dx
        else:
            return avg_f
    else:
        return avg_f


def integrate_discrete_squared_curvature(
        f: Union[_onp.ndarray, _cas.MX],
        x: Union[_onp.ndarray, _cas.MX] = None,
        method: str = "hybrid_simpson_cubic",
):
    """
    Given a set of sampled points (x_i, f_i) from a function f(x), computes the following quantity:

        int_{x[0]}^{x[-1]} (f''(x))^2 dx

    This is useful for regularization of smooth curves (i.e., encouraging smooth functions as optimization results).

    Performs this through one of several reconstruction-based methods, specified by `method`:

        * "cubic": On each interval, reconstructs a piecewise cubic polynomial. This cubic is the unique polynomial
        that passes through the two points at the endpoints of the interval, plus the next point beyond each endpoint
        of the interval (i.e., 4 points in total). Numerically, this cubic is obtained using Bernstein polynomial
        reconstruction, so it is numerically stable. This cubic is then analytically differentiated twice, squared,
        and integrated over the interval. At the ends of the overall array, where this "look beyond" strategy is not
        possible, a one-sided cubic is used instead (i.e., looks beyond the interval at one end only and uses two
        extra points from this side).

        * "simpson": On each interval, makes two unique quadratic reconstructions:

            * One reconstruction that uses the two points at the endpoints of the interval, plus the next point beyond
            the right endpoint of the interval (i.e., 3 points in total).

            * One reconstruction that uses the two points at the endpoints of the interval, plus the next point beyond
            the left endpoint of the interval (i.e., 3 points in total).

            These two quadratics are then analytically differentiated twice, squared, and integrated over the
            interval. This requires much less calculation, since the quadratics have uniform curvature over the
            interval, causing a lot of things to simplify. The result is then computed by combining the results of this
            process for the two quadratic reconstructions.

            This is similar to a Simpson's rule integration, balanced between the two sides of the interval. In
            frequency-domain testing, this method appears to be more accurate than the "cubic" strategy at every
            frequency, with less computational effort. Thus, it should be preferred to the "cubic" strategy.

        * "hybrid_simpson_cubic": First, starts out by estimating the first derivative of the function at each point
        in the array (including endpoints) using a quadratic reconstruction. (See `numpy.gradient()` for more
        information or source code on this; this code uses `numpy.gradient()` directly for this step.) Then,
        reconstructs a cubic polynomial on each interval, with the following boundary conditions:

            * The cubic passes through the two points at the endpoints of the interval.

            * The cubic has the same first derivative as the precomputed derivatives at the endpoints of the interval.

            This cubic is then analytically differentiated twice, squared, and integrated over the interval.

            In frequency-domain testing, this method is also more accurate than the "cubic" strategy at every
            frequency. Compared to the "simpson" strategy, it is more accurate at high frequencies and less accurate
            at low frequencies. Because the goal of this function is to be used as a regularization term,
            which should be more sensitive to high-frequency oscillations, this method is preferred to the "simpson"
            strategy. This method is also preferred as its estimate tends to err high rather than low, which serves
            well as a regularization strategy. (It is still convergent to the true value in the high-sample-rate limit.)

    """
    # Determine if an x-array was specified, and calculate dx.
    x_is_specified = (x is not None)
    if not x_is_specified:
        x = _onp.arange(length(f))

    if method in ["cubic", "cubic_spline"]:
        x1 = x[:-3]
        x2 = x[1:-2]
        x3 = x[2:-1]
        x4 = x[3:]

        f1 = f[:-3]
        f2 = f[1:-2]
        f3 = f[2:-1]
        f4 = f[3:]

        h = x3 - x2
        hm = x2 - x1
        hp = x4 - x3

        dfm = f2 - f1
        df = f3 - f2
        dfp = f4 - f3

        ### The following section computes the integral of the squared second derivative of the cubic spline interpolant
        ### for the "middle" intervals (i.e. not the first or last intervals).
        ### Code is generated by sympy; here s_i variables represent common subexpressions.
        s0 = hm ** 2
        s1 = hp ** 2
        s2 = h + hm
        s3 = h ** 2
        s4 = hp ** 6
        s5 = h ** 6
        s6 = hp ** 5
        s7 = h ** 3
        s8 = 3 * s7
        s9 = hp ** 4
        s10 = h ** 4
        s11 = 4 * s10
        s12 = hp ** 3
        s13 = 3 * h ** 5
        s14 = hm ** 6
        s15 = hm ** 5
        s16 = hm ** 4
        s17 = hm ** 3
        s18 = hm * s10 * s12
        s19 = hp * s10 * s17
        s20 = s12 * s17
        s21 = s0 * s12 * s8
        s22 = s1 * s17 * s8
        s23 = 3 * s0 * s1 * s10 + s21 + s22
        s24 = 2 * h
        s25 = 6 * s3
        s26 = 7 * s7
        s27 = 3 * s3
        s28 = -s20 * s27
        s29 = 3 * h
        middle_intervals = 4 * (df ** 2 * (
                s0 * s27 * s9 + s0 * s29 * s6 + s0 * s4 + s1 * s14 + s1 * s15 * s29 + s1 * s16 * s27 - s12 * s16 * s29 - 2 * s16 * s9 - s17 * s29 * s9 + s23 + s28) + df * dfm * (
                                        2 * h * s17 * s9 - hm * s1 * s13 - hm * s24 * s4 - hm * s25 * s6 - hm * s26 * s9 + 3 * s0 * s3 * s9 + s1 * s17 * s7 - 6 * s18 + s21 - s28) + df * dfp * (
                                        2 * h * s12 * s16 - hp * s0 * s13 - hp * s14 * s24 - hp * s15 * s25 - hp * s16 * s26 + s0 * s12 * s7 + 3 * s1 * s16 * s3 - 6 * s19 + s22 - s28) + dfm ** 2 * (
                                        s1 * s5 + s11 * s9 + s12 * s13 + s3 * s4 + s6 * s8) + dfm * dfp * (
                                        hm * hp * s5 - s18 - s19 - 2 * s20 * s3 - s23) + dfp ** 2 * (
                                        s0 * s5 + s11 * s16 + s13 * s17 + s14 * s3 + s15 * s8)) / (
                                   h * s0 * s1 * s2 ** 2 * (h + hp) ** 2 * (hp + s2) ** 2)

        ### Now we compute the integral for the first interval.
        h_f = h[slice(0, 1)]
        hm_f = hm[slice(0, 1)]
        hp_f = hp[slice(0, 1)]
        df_f = df[slice(0, 1)]
        dfm_f = dfm[slice(0, 1)]
        dfp_f = dfp[slice(0, 1)]

        s0 = h_f ** 2
        s1 = hp_f ** 2
        s2 = h_f + hm_f
        s3 = hp_f ** 6
        s4 = df_f * dfm_f
        s5 = hm_f ** 6
        s6 = 2 * dfp_f
        s7 = df_f * s6
        s8 = h_f ** 6
        s9 = dfm_f * dfp_f
        s10 = 4 * s9
        s11 = df_f ** 2
        s12 = hm_f ** 2
        s13 = dfm_f ** 2
        s14 = dfp_f ** 2
        s15 = hm_f ** 3
        s16 = hp_f ** 5
        s17 = 3 * s11
        s18 = hp_f ** 4
        s19 = hm_f ** 4
        s20 = 4 * s19
        s21 = hm_f ** 5
        s22 = hp_f ** 3
        s23 = h_f ** 3
        s24 = 6 * s13
        s25 = h_f ** 4
        s26 = h_f ** 5
        s27 = 3 * s14
        s28 = df_f * dfp_f
        s29 = -dfm_f
        s30 = df_f * h_f
        s31 = 15 * df_f
        s32 = 2 * df_f
        s33 = dfm_f * hm_f
        s34 = 3 * dfm_f
        s35 = -s34
        s36 = dfp_f * hp_f
        s37 = 3 * s12
        s38 = 5 * s11
        s39 = 3 * s0
        s40 = -3 * s28 + s9
        s41 = 18 * s11
        s42 = -s7
        first_interval = 4 * (
                -2 * h_f * hm_f * s3 * s4 - h_f * hp_f * s5 * s7 + hm_f * hp_f * s10 * s8 - hm_f * s0 * s16 * s34 * (
                4 * df_f + s29) - 9 * hp_f * s0 * s21 * s28 + s0 * s13 * s3 + s0 * s14 * s5 + s0 * s15 * s22 * (
                        27 * s11 - 21 * s4 + s40) + s1 * s11 * s5 + 4 * s1 * s13 * s8 + s1 * s15 * s23 * (
                        -12 * s28 - 14 * s4 + s41 + 9 * s9) + s1 * s19 * s39 * (
                        s38 - s4 + s40) + 3 * s1 * s21 * s30 * (-dfp_f + s32) + s1 * s25 * s37 * (
                        s10 + s13 + s17 - 7 * s4 + s42) + 6 * s1 * s26 * s33 * (
                        dfm_f + dfp_f - s32) + s11 * s12 * s3 + s11 * s18 * s20 + s12 * s14 * s8 + 6 * s12 * s16 * s30 * (
                        df_f + s29) + s12 * s18 * s39 * (s13 + s38 - 9 * s4) + s12 * s22 * s23 * (
                        s24 - 42 * s4 + s41 + s42 + 3 * s9) + 13 * s13 * s18 * s25 + 12 * s13 * s22 * s26 + s14 * s20 * s25 + s15 * s16 * s17 + s15 * s18 * s30 * (
                        -7 * dfm_f + s31) - s15 * s25 * s36 * (
                        -8 * dfm_f + s31) + s15 * s26 * s27 + s16 * s23 * s24 + s17 * s21 * s22 - 4 * s18 * s23 * s33 * (
                        7 * df_f + s35) - s19 * s22 * s30 * (dfp_f - s31 + s34) - s19 * s23 * s36 * (
                        16 * df_f + s35) + s21 * s23 * s27 + s22 * s25 * s33 * (
                        -30 * df_f + 15 * dfm_f + s6) - s26 * s36 * s37 * (s32 + s35)) / (
                                 hm_f * s0 * s1 * s2 ** 2 * (h_f + hp_f) ** 2 * (hp_f + s2) ** 2)

        ### Now we compute the integral for the last interval.
        h_l = h[slice(-1, None)]
        hm_l = hm[slice(-1, None)]
        hp_l = hp[slice(-1, None)]
        df_l = df[slice(-1, None)]
        dfm_l = dfm[slice(-1, None)]
        dfp_l = dfp[slice(-1, None)]

        s0 = h_l ** 2
        s1 = hm_l ** 2
        s2 = h_l + hm_l
        s3 = hp_l ** 6
        s4 = 2 * dfm_l
        s5 = df_l * s4
        s6 = hm_l ** 6
        s7 = df_l * dfp_l
        s8 = h_l ** 6
        s9 = dfm_l * dfp_l
        s10 = 4 * s9
        s11 = df_l ** 2
        s12 = hp_l ** 2
        s13 = dfm_l ** 2
        s14 = dfp_l ** 2
        s15 = hm_l ** 3
        s16 = hp_l ** 5
        s17 = 3 * s11
        s18 = hm_l ** 4
        s19 = hp_l ** 4
        s20 = 4 * s19
        s21 = hm_l ** 5
        s22 = hp_l ** 3
        s23 = h_l ** 3
        s24 = 3 * s13
        s25 = h_l ** 4
        s26 = h_l ** 5
        s27 = 6 * s14
        s28 = df_l * dfm_l
        s29 = -dfp_l
        s30 = df_l * h_l
        s31 = 15 * df_l
        s32 = 2 * df_l
        s33 = dfp_l * hp_l
        s34 = 3 * dfp_l
        s35 = -s34
        s36 = dfm_l * hm_l
        s37 = 3 * s12
        s38 = 5 * s11
        s39 = 3 * s0
        s40 = -3 * s28 + s9
        s41 = 18 * s11
        s42 = -s5
        last_interval = 4 * (
                -h_l * hm_l * s3 * s5 - 2 * h_l * hp_l * s6 * s7 + hm_l * hp_l * s10 * s8 - 9 * hm_l * s0 * s16 * s28 - hp_l * s0 * s21 * s34 * (
                4 * df_l + s29) + s0 * s13 * s3 + s0 * s14 * s6 + s0 * s15 * s22 * (
                        27 * s11 + s40 - 21 * s7) + s1 * s11 * s3 + 4 * s1 * s14 * s8 + 3 * s1 * s16 * s30 * (
                        -dfm_l + s32) + s1 * s19 * s39 * (s38 + s40 - s7) + s1 * s22 * s23 * (
                        -12 * s28 + s41 - 14 * s7 + 9 * s9) + s1 * s25 * s37 * (
                        s10 + s14 + s17 + s42 - 7 * s7) + 6 * s1 * s26 * s33 * (
                        dfm_l + dfp_l - s32) + s11 * s12 * s6 + s11 * s18 * s20 + s12 * s13 * s8 + s12 * s15 * s23 * (
                        s27 + s41 + s42 - 42 * s7 + 3 * s9) + s12 * s18 * s39 * (
                        s14 + s38 - 9 * s7) + 6 * s12 * s21 * s30 * (
                        df_l + s29) + s13 * s20 * s25 + 12 * s14 * s15 * s26 + 13 * s14 * s18 * s25 + s15 * s16 * s17 - s15 * s19 * s30 * (
                        dfm_l - s31 + s34) + s15 * s25 * s33 * (
                        -30 * df_l + 15 * dfp_l + s4) + s16 * s23 * s24 + s17 * s21 * s22 + s18 * s22 * s30 * (
                        -7 * dfp_l + s31) - 4 * s18 * s23 * s33 * (7 * df_l + s35) - s19 * s23 * s36 * (
                        16 * df_l + s35) + s21 * s23 * s27 + s22 * s24 * s26 - s22 * s25 * s36 * (
                        -8 * dfp_l + s31) - s26 * s36 * s37 * (s32 + s35)) / (
                                hp_l * s0 * s1 * s2 ** 2 * (h_l + hp_l) ** 2 * (hp_l + s2) ** 2)

        ### Now, we stitch together the intervals.
        res = concatenate((
            first_interval,
            middle_intervals,
            last_interval,
        ))

        return res

    elif method in ["simpson"]:
        ### Forward Simpson for intervals 0 to N-2
        x2 = x[:-2]
        x3 = x[1:-1]
        x4 = x[2:]

        f2 = f[:-2]
        f3 = f[1:-1]
        f4 = f[2:]

        h = x3 - x2
        hp = x4 - x3

        df = f3 - f2
        dfp = f4 - f3

        res_forward_simpson = 4 * (df * hp - dfp * h) ** 2 / (h * hp ** 2 * (h + hp) ** 2)

        ### Backward Simpson for intervals 1 to N-1
        x1 = x[:-2]
        x2 = x[1:-1]
        x3 = x[2:]

        f1 = f[:-2]
        f2 = f[1:-1]
        f3 = f[2:]

        h = x3 - x2
        hm = x2 - x1

        dfm = f2 - f1
        df = f3 - f2

        res_backward_simpson = 4 * (df * hm - dfm * h) ** 2 / (h * hm ** 2 * (h + hm) ** 2)

        ### Fuse them together
        first_interval = res_forward_simpson[slice(0, 1)]

        a = res_backward_simpson[slice(None, -1)]
        b = res_forward_simpson[slice(1, None)]

        # middle_intervals = (a + b) / 2
        middle_intervals = ((a ** 2 + b ** 2) / 2 + 1e-100) ** 0.5  # This is more accurate across all frequencies

        last_interval = res_backward_simpson[slice(-1, None)]

        res = concatenate((
            first_interval,
            middle_intervals,
            last_interval,
        ))

        return res

    elif method in ["hybrid_simpson_cubic"]:
        from aerosandbox.numpy.calculus import gradient
        dfdx = gradient(
            f,
            x,
            edge_order=2
        )

        h = x[1:] - x[:-1]
        df = f[1:] - f[:-1]
        dfdx1 = dfdx[:-1]
        dfdx2 = dfdx[1:]

        res = (
                4 * (dfdx1 ** 2 + dfdx1 * dfdx2 + dfdx2 ** 2) / h
                + 12 * df / h ** 2 * (df / h - dfdx1 - dfdx2)
        )

        return res


    else:
        raise ValueError(f"Invalid method '{method}'.")


if __name__ == '__main__':
    import aerosandbox as asb
    import aerosandbox.numpy as np
    from scipy import integrate, interpolate
    import sympy as s

    np.random.seed(0)

    # degree = 4
    # coeffs = np.random.randn(degree + 1)
    # a = 1
    # b = 3
    #
    #
    # def f(x):
    #     out = 0
    #     for i in range(degree + 1):
    #         out += coeffs[i] * x ** i
    #
    #     return out

    a = 0
    b = 2


    def f(x):
        sin = np.sin if isinstance(x, np.ndarray) else s.sin
        return sin(2 * np.pi * x * 1) + 1


    print("\n\nTest 1: Integration")
    exact = integrate.quad(
        f,
        a,
        b,
        epsrel=1e-15,
    )[0]

    x_vals = np.cosspace(a, b, 100)
    f_vals = f(x_vals)

    approx_intervals = integrate_discrete_intervals(
        f=f_vals,
        x=x_vals,
        multiply_by_dx=True,
        # method="trapz",
        method="cubic",
        # method_endpoints="ignore",
    )

    integral = np.sum(approx_intervals)

    print(f"exact: {exact}")
    print(f"approx: {integral}")
    print(f"error: {integral - exact}")

    print("\n\nTest 2: Squared curvature")

    x_sym = s.symbols("x")
    f_sym = f(x_sym)
    df2dx_func = s.lambdify(x_sym, s.diff(f_sym, x_sym, 2))

    exact = integrate.quad(
        lambda x: df2dx_func(x) ** 2,
        a,
        b,
        epsrel=1e-15,
    )[0]
    print(f"exact: {exact}")

    approx = integrate_discrete_squared_curvature(
        f=f_vals,
        x=x_vals,
        # method="simpson"
        method="hybrid_simpson_cubic",
    )
    integral = np.sum(approx)
    print(f"\nintegrate_discrete_squared_curvature: {integral}")
    print(f"error: {integral - exact}")

    approx = integrate_discrete_intervals(
        f=np.gradient(
            f_vals,
            x_vals,
            n=2
        ) ** 2,
        x=x_vals,
    )
    integral = np.sum(approx)

    print(f"\nintegrate_discrete_intervals + np.gradient: {integral}")
    print(f"error: {integral - exact}")

    print("\n\nTest 3: Squared curvature with global-spline reconstruction")

    x = np.arange(0, 100 + 1)
    f = np.cos(np.pi * x / 2)

    f_interp = interpolate.InterpolatedUnivariateSpline(
        x=x,
        y=f,
        k=3
    )
    exact = integrate.quad(
        lambda x: f_interp.derivative(2)(x) ** 2,
        x[0],
        x[-1],
        epsrel=1e-8,
    )[0]

    approx = integrate_discrete_squared_curvature(
        f=f,
        x=x,
        method="hybrid_simpson_cubic",
        # method="simpson"
    )
    integral = np.sum(approx)

    print(f"exact: {exact}")
    print(f"\nintegrate_discrete_squared_curvature: {integral}")
    print(f"error: {integral - exact}")

    x_plot = np.linspace(x[0], x[-1], 10000)
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    # p.qp(x_plot, f_interp.derivative(2)(x_plot) ** 2)



"""
Simple Beam

A simple 2D beam example, to be integrated later for full aerostructural modeling. TODO do that.

Governing equation for bending:
Euler-Bernoulli beam theory.

(E * I * u(x)'')'' = q(x)

where:
    * E is the elastic modulus
    * I is the bending moment of inertia
    * u(x) is the local displacement at x.
    * q(x) is the force-per-unit-length at x. (In other words, a dirac delta is a point load.)
    * ()' is a derivative w.r.t. x.

Governing equation for torsion:
phi(x)'' = -T / (G * J)

where:
    * phi is the local twist angle
    * T is the local torque per unit length
    * G is the local shear modulus
    * J is the polar moment of inertia
    * ()' is a derivative w.r.t. x.

"""
import aerosandbox.numpy as np
import casadi as cas

if __name__ == '__main__':

    opti = cas.Opti()  # Initialize a SAND environment

    # Define Assumptions
    L = 34.1376 / 2
    n = 50
    mass_total = 292
    x = cas.linspace(0, L, n)
    dx = cas.diff(x)
    E = 228e9  # Pa, modulus of CF
    G = E / 2 / (1 + 0.5)  # TODO fix this!!! CFRP is not isotropic!
    max_allowable_stress = 570e6 / 1.75

    log_nominal_diameter = opti.variable(n)
    opti.set_initial(log_nominal_diameter, cas.log(200e-3))
    nominal_diameter = cas.exp(log_nominal_diameter)

    thickness = 0.14e-3 * 5
    opti.subject_to([
        nominal_diameter > thickness,
    ])


    def trapz(x):
        out = (x[:-1] + x[1:]) / 2
        out[0] += x[0] / 2
        out[-1] += x[-1] / 2
        return out


    # Mass
    volume = cas.sum1(
        cas.pi / 4 * trapz((nominal_diameter + thickness) ** 2 - (nominal_diameter - thickness) ** 2) * dx
    )
    mass = volume * 1600

    # Bending loads
    I = cas.pi / 64 * ((nominal_diameter + thickness) ** 4 - (nominal_diameter - thickness) ** 4)
    EI = E * I
    total_lift_force = 9.81 * (mass_total - mass) / 2  # 9.81 * 292 / 2
    lift_distribution = "elliptical"
    if lift_distribution == "rectangular":
        force_per_unit_length = total_lift_force * cas.GenDM_ones(n) / L
    elif lift_distribution == "elliptical":
        force_per_unit_length = total_lift_force * cas.sqrt(1 - (x / L) ** 2) * (4 / cas.pi) / L

    # Torsion loads
    J = cas.pi / 32 * ((nominal_diameter + thickness) ** 4 - (nominal_diameter - thickness) ** 4)

    airfoil_lift_coefficient = 1
    airfoil_moment_coefficient = -0.14
    airfoil_chord = 1  # meter
    moment_per_unit_length = force_per_unit_length * airfoil_moment_coefficient * airfoil_chord / airfoil_lift_coefficient
    # Derivation of above:
    #   CL = L / q c
    #   CM = M / q c**2
    #   M / L = (CM * c) / (CL)

    # Set up derivatives
    u = 1 * opti.variable(n)
    du = 0.1 * opti.variable(n)
    ddu = 0.01 * opti.variable(n)
    dEIddu = 100 * opti.variable(n)
    phi = 0.1 * opti.variable(n)
    dphi = 0.01 * opti.variable(n)
    # opti.set_initial(u, 2 * (x/L)**4)
    # opti.set_initial(du, 2 * 4/L * (x/L)**3)
    # opti.set_initial(ddu, 2 * 3/L * 2/L * (x/L))
    # opti.set_initial(dEIddu, 2 * 3/L * 2/L * 1/L * 1e3)

    # Add forcing term
    ddEIddu = force_per_unit_length
    ddphi = -moment_per_unit_length / (G * J)

    # Define derivatives
    opti.subject_to([
        cas.diff(u) == trapz(du) * dx,
        cas.diff(du) == trapz(ddu) * dx,
        cas.diff(EI * ddu) == trapz(dEIddu) * dx,
        cas.diff(dEIddu) == trapz(ddEIddu) * dx,
        cas.diff(phi) == trapz(dphi) * dx,
        cas.diff(dphi) == trapz(ddphi) * dx,
    ])

    # Add BCs
    opti.subject_to([
        u[0] == 0,
        du[0] == 0,
        ddu[-1] == 0,  # No tip moment
        dEIddu[-1] == 0,  # No tip higher order stuff
        phi[0] == 0,
        dphi[-1] == 0,
    ])

    # Failure criterion
    stress_axial = (nominal_diameter + thickness) / 2 * E * ddu
    stress_shear = dphi * G * (nominal_diameter + thickness) / 2
    # stress_axial = cas.fmax(0, stress_axial)
    # stress_shear = cas.fmax(0, stress_shear)
    stress_von_mises_squared = cas.sqrt(
        stress_axial ** 2 + 0 * stress_shear ** 2)  # Source: https://en.wikipedia.org/wiki/Von_Mises_yield_criterion
    stress = stress_axial
    opti.subject_to([
        stress / max_allowable_stress < 1
    ])

    opti.minimize(mass)

    # Tip deflection constraint
    opti.subject_to([
        # u[-1] < 2 # Source: http://web.mit.edu/drela/Public/web/hpa/hpa_structure.pdf
        du[-1] * 180 / cas.pi < 10
    ])

    # Twist
    opti.subject_to([
        phi[-1] * 180 / cas.pi > -5
    ])

    p_opts = {}
    s_opts = {}
    s_opts["max_iter"] = 500  # If you need to interrupt, just use ctrl+c
    # s_opts["mu_strategy"] = "adaptive"
    opti.solver('ipopt', p_opts, s_opts)

    try:
        sol = opti.solve()
    except Exception:
        print("Failed!")
        sol = opti.debug

    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set(font_scale=1)

    fig, ax = plt.subplots(2, 3, figsize=(10, 6), dpi=200)

    plt.subplot(231)
    plt.plot(sol(x), sol(u), '.-')
    plt.xlabel("x [m]")
    plt.ylabel("u [m]")
    plt.title("Displacement (Bending)")
    plt.axis("equal")

    # plt.subplot(232)
    # plt.plot(sol.value(x), np.arctan(sol.value(du))*180/np.pi, '.-')
    # plt.xlabel("x [m]")
    # plt.ylabel(r"Local Slope [deg]")
    # plt.title("Slope")

    plt.subplot(232)
    plt.plot(sol.value(x), sol.value(phi) * 180 / np.pi, '.-')
    plt.xlabel("x [m]")
    plt.ylabel("Twist angle [deg]")
    plt.title("Twist Angle (Torsion)")

    plt.subplot(233)
    plt.plot(sol.value(x), sol.value(force_per_unit_length), '.-')
    plt.xlabel("x [m]")
    plt.ylabel(r"$F$ [N/m]")
    plt.title("Local Load per Unit Span")

    plt.subplot(234)
    plt.plot(sol.value(x), sol.value(stress / 1e6), '.-')
    plt.xlabel("x [m]")
    plt.ylabel("Stress [MPa]")
    plt.title("Peak Stress at Section")

    plt.subplot(235)
    plt.plot(sol.value(x), sol.value(dEIddu), '.-')
    plt.xlabel("x [m]")
    plt.ylabel("F [N]")
    plt.title("Shear Force")

    plt.subplot(236)
    plt.plot(sol.value(x), sol.value(nominal_diameter), '.-')
    plt.xlabel("x [m]")
    plt.ylabel("t [m]")
    plt.title("Optimal Spar Diameter")

    plt.suptitle(f"Beam Modeling (Total Spar Mass: {2 * sol.value(mass):.2f} kg)")

    plt.subplots_adjust(hspace=0.4)

    # plt.tight_layout()
    # plt.legend()
    plt.show()

    print("Mass (half-wing) [kg]:", sol.value(mass))
    print("Mass (full-wing) [kg]:", 2 * sol.value(mass))



import aerosandbox as asb
import aerosandbox.numpy as np
import aerosandbox.tools.units as u


def tire_size(
        mass_supported_by_each_tire: float,
        aircraft_type="general_aviation"
) -> float:
    """
    Computes the required diameter and width of a tire for an airplane, from statistical regression to historical data.

    Methodology and constants from Raymer: Aircraft Design: A Conceptual Approach, 5th Edition, Table 11.1, pg. 358.

    Args:
        mass_supported_by_each_tire: The mass supported by each tire, in kg.

        aircraft_type: The type of aircraft. Options are:
            - "general_aviation"
            - "business_twin"
            - "transport/bomber"
            - "fighter/trainer"

    Returns:
        The required diameter and width of the tire, in meters.
    """
    mass_supported_by_tire_lbm = mass_supported_by_each_tire / u.lbm

    if aircraft_type == "general_aviation":
        A = 1.51
        B = 0.349
    elif aircraft_type == "business_twin":
        A = 2.69
        B = 0.251
    elif aircraft_type == "transport/bomber":
        A = 1.63
        B = 0.315
    elif aircraft_type == "fighter/trainer":
        A = 1.59
        B = 0.302
    else:
        raise ValueError("Invalid `aircraft_type`.")

    tire_diameter_in = A * mass_supported_by_tire_lbm ** B

    if aircraft_type == "general_aviation":
        A = 0.7150
        B = 0.312
    elif aircraft_type == "business_twin":
        A = 1.170
        B = 0.216
    elif aircraft_type == "transport/bomber":
        A = 0.1043
        B = 0.480
    elif aircraft_type == "fighter/trainer":
        A = 0.0980
        B = 0.467

    tire_width_in = A * mass_supported_by_tire_lbm ** B

    tire_diameter = tire_diameter_in * u.inch
    tire_width = tire_width_in * u.inch

    return tire_diameter, tire_width



import sympy as s
from sympy import init_printing

init_printing()

def simplify(expr, maxdepth=10, _depth=0):
    import copy
    original_expr = copy.copy(expr)
    print(f"Depth: {_depth} | Parsimony: {len(str(expr))}")
    expr = expr.simplify()
    if expr != original_expr:
        if len(str(expr)) < len(str(original_expr)):
            if _depth < maxdepth:
                return simplify(expr, maxdepth=maxdepth, _depth=_depth + 1)
            else:
                return expr
        else:
            return original_expr
    else:
        return expr
##### Now, a new problem:
##### Do a cubic reconstruction of an interval using values and first derivatives at the endpoints.

# x1, x2 = s.symbols("x1 x2", real=True)
h = s.symbols("h", real=True)

x1 = 0
x2 = h

f1, f2 = s.symbols("f1 f2", real=True)
dfdx1, dfdx2 = s.symbols("dfdx1 dfdx2", real=True)

q = s.symbols("q", real=True)

q1 = 0
q2 = 1
dqdx = 1 / (x2 - x1)

# Define the Bernstein basis polynomials
b1 = (1 - q) ** 3
b2 = 3 * q * (1 - q) ** 2
b3 = 3 * q ** 2 * (1 - q)
b4 = q ** 3

c1, c2, c3, c4 = s.symbols('c1 c2 c3 c4', real=True)

# Can solve for c1 and c4 exactly
c1 = f1
c4 = f2

f = c1 * b1 + c2 * b2 + c3 * b3 + c4 * b4
dfdx = f.diff(q) * dqdx
# Solve for c2 and c3
sol = s.solve(
    [
        dfdx.subs(q, q1) - dfdx1,
        dfdx.subs(q, q2) - dfdx2,
    ],
    [
        c2,
        c3,
    ],
)
c2_sol = simplify(sol[c2])
c3_sol = simplify(sol[c3])

f = c1 * b1 + c2_sol * b2 + c3_sol * b3 + c4 * b4

dfdx = f.diff(q) * dqdx
df2dx = dfdx.diff(q) * dqdx

res = s.integrate(df2dx ** 2, (q, 0, 1)) * h
res = simplify(res)

cse = s.cse(
    res,
    symbols=s.numbered_symbols("s"),
    list=False
)

parsimony = len(str(res))
print("\nSquared Curvature:")
print(f"Parsimony: {parsimony}")

for i, (var, expr) in enumerate(cse[0]):
    print(f"{var} = {expr}")
print(f"res = {cse[1]}")



import sympy as s
from sympy import init_printing

init_printing()

# Gets the integral from x2 to x3, looking at the cubic spline interpolant from x1...x4

# Define the symbols
hm, h, hp = s.symbols("hm h hp", real=True)
dfm, df, dfp = s.symbols("dfm df dfp", real=True)

x1, x2, x3, x4 = s.symbols('x1 x2 x3 x4', real=True)
# f1, f2, f3, f4 = s.symbols('f1 f2 f3 f4', real=True)

f1 = 0
f2 = dfm
f3 = f2 + df
f4 = f3 + dfp


def simplify(expr, maxdepth=10, _depth=0):
    import copy
    original_expr = copy.copy(expr)
    print(f"Depth: {_depth} | Parsimony: {len(str(expr))}")
    maps = {
        f4 - f3: dfp,
        f3 - f2: df,
        f2 - f1: dfm,
        f4 - f2: dfp + df,
        f3 - f1: df + dfm,
        f4 - f1: dfp + df + dfm,
        x4 - x3: hp,
        x3 - x2: h,
        x2 - x1: hm,
        x4 - x2: hp + h,
        x3 - x1: h + hm,
        x4 - x1: hp + h + hm,
    }
    expr = expr.factor([h, hm, hp])
    expr = expr.subs(maps)
    expr = expr.simplify()
    if expr != original_expr:
        if len(str(expr)) < len(str(original_expr)):
            if _depth < maxdepth:
                return simplify(expr, maxdepth=maxdepth, _depth=_depth + 1)
            else:
                return expr
        else:
            return original_expr
    else:
        return expr


q = s.symbols('q')  # Normalized space for a Bernstein basis.
# Mapping from x-space to q-space has x=x2 -> q=0, x=x3 -> q=1.

q2 = 0
q3 = 1
q1 = q2 - hm / h
q4 = q3 + hp / h
# q1, q4 = s.symbols('q1 q4', real=True)

# Define the Bernstein basis polynomials
b1 = (1 - q) ** 2
b2 = 2 * q * (1 - q)
b3 = q ** 2

c1, c2, c3, c4 = s.symbols('c1 c2 c3 c4', real=True)

# Can solve for c2 and c3 exactly
c1 = f2
c3 = f3

f = c1 * b1 + c2 * b2 + c3 * b3

f1_cubic = f.subs(q, q1)  # .simplify()
# f4_cubic = f.subs(q, q4)  # .simplify()

# factors = [q1, q4]
# factors = [f1, f2, f3, f4]

# Solve for c2 and c3
sol = s.solve(
    [
        f1_cubic - f1,
        # f4_cubic - f4,
    ],
    [
        c2,
    ],
)
c2_sol = simplify(sol[c2].factor([dfm, df, dfp]))

dfdx = f.diff(q) / h
df2dx = dfdx.diff(q) / h

res = s.integrate(df2dx ** 2, (q, q2, q3)) * h
res = simplify(res.factor([dfm, df, dfp]))
res = simplify(res)

res = simplify(res.subs({c2: c2_sol}).factor([dfm, df, dfp]))

cse = s.cse(
    res,
    symbols=s.numbered_symbols("s"),
    list=False)

parsimony = len(str(res))
# print(s.pretty(res, num_columns=100))
print(f"Parsimony: {parsimony}")

for i, (var, expr) in enumerate(cse[0]):
    print(f"{var} = {expr}")
print(f"res = {cse[1]}")

if __name__ == '__main__':
    x = s.symbols('x', real=True)

    a = 0
    b = 4

    def example_f(x):
        return x ** 2 + 2 * x + 3

    h_val = b - a
    hm_val = 1
    hp_val = 3
    df_val = example_f(b) - example_f(a)
    dfm_val = example_f(a) - example_f(a - hm_val)
    dfp_val = example_f(b + hp_val) - example_f(b)

    subs = {
            h: h_val,
            hm: hm_val,
            hp: hp_val,
            df: df_val,
            dfm: dfm_val,
            dfp: dfp_val,
        }

    exact = s.N(
        s.integrate(
            example_f(x).diff(x, 2) ** 2,
            (x, a, b)
        )
    )

    eqn = s.N(
        res.subs(subs)
    )

    print(f"exact: {exact}")
    print(f"eqn: {eqn}")
    print(f"ratio: {exact/eqn}")


import aerosandbox as asb
import aerosandbox.numpy as np
import pytest

airplane = asb.Airplane(
    wings=[
        asb.Wing(
            xsecs=[
                asb.WingXSec(
                    xyz_le=[0, 0, 0],
                    chord=1,
                    airfoil=asb.Airfoil("naca0001")
                ),
                asb.WingXSec(
                    xyz_le=[0, 1, 0],
                    chord=1,
                    airfoil=asb.Airfoil("naca0001")
                )
            ]
        )
    ]
)


def LD_from_alpha(alpha):
    op_point = asb.OperatingPoint(
        velocity=1,
        alpha=alpha,
    )

    vlm = asb.VortexLatticeMethod(
        airplane,
        op_point,
        align_trailing_vortices_with_wind=True
    )
    aero = vlm.run()

    CD0 = 0.01

    LD = aero["CL"] / (aero["CD"] + CD0)
    return LD


def test_vlm_optimization_operating_point():
    opti = asb.Opti()
    alpha = opti.variable(init_guess=0, lower_bound=-30, upper_bound=30)
    LD = LD_from_alpha(alpha)
    opti.minimize(-LD)
    sol = opti.solve(
        verbose=True,
        # callback=lambda _: print(f"alpha = {opti.debug.value(alpha)}")
    )
    print(sol(alpha), sol(LD))
    assert sol(alpha) == pytest.approx(5.85, abs=0.1)


if __name__ == '__main__':
    LD_from_alpha(6)
    test_vlm_optimization_operating_point()
    # pytest.main()



from aerosandbox import AeroSandboxObject
from aerosandbox.geometry.common import *
from typing import List, Dict, Any, Union, Tuple, Optional
import copy


class Propulsor(AeroSandboxObject):
    """
    Definition for a Propulsor, which could be a propeller, a rotor, or a jet engine.

    Assumes a disk- or cylinder-shaped propulsor.
    """

    def __init__(self,
                 name: Optional[str] = "Untitled",
                 xyz_c: Union[np.ndarray, List[float]] = None,
                 xyz_normal: Union[np.ndarray, List[float]] = None,
                 radius: float = 1.,
                 length: float = 0.,
                 color: Optional[Union[str, Tuple[float]]] = None,
                 analysis_specific_options: Optional[Dict[type, Dict[str, Any]]] = None,
                 ):
        """
        Defines a new propulsor object.

        TODO add docs
        """
        ### Set defaults
        if xyz_c is None:
            xyz_c = np.array([0., 0., 0.])
        if xyz_normal is None:
            xyz_normal = np.array([-1., 0., 0.])
        if analysis_specific_options is None:
            analysis_specific_options = {}

        self.name = name
        self.xyz_c = np.array(xyz_c)
        self.xyz_normal = np.array(xyz_normal)
        self.radius = radius
        self.length = length
        self.color = color
        self.analysis_specific_options = analysis_specific_options

    def __repr__(self) -> str:
        return f"Propulsor '{self.name}' (xyz_c: {self.xyz_c}, radius: {self.radius})"

    def xsec_area(self) -> float:
        """
        Returns the cross-sectional area of the propulsor, in m^2.
        """
        return np.pi * self.radius ** 2

    def xsec_perimeter(self) -> float:
        """
        Returns the cross-sectional perimeter of the propulsor, in m.
        """
        return 2 * np.pi * self.radius

    def volume(self) -> float:
        """
        Returns the volume of the propulsor, in m^3.
        """
        return self.xsec_area() * self.length

    def compute_frame(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes the local coordinate frame of the propulsor, in aircraft geometry axes.

        xg_local is aligned with the propulsor's normal vector.

        zg_local is roughly aligned with the z-axis of the aircraft geometry axes, but projected onto the propulsor's plane.

        yg_local is the cross product of zg_local and xg_local.

        Returns: A tuple:
            xg_local: The x-axis of the local coordinate frame, in aircraft geometry axes.
            yg_local: The y-axis of the local coordinate frame, in aircraft geometry axes.
            zg_local: The z-axis of the local coordinate frame, in aircraft geometry axes.

        """
        xyz_normal = self.xyz_normal / np.linalg.norm(self.xyz_normal)

        xg_local = xyz_normal

        zg_local = np.array([0, 0, 1])
        zg_local = zg_local - np.dot(zg_local, xg_local) * xg_local

        yg_local = np.cross(zg_local, xg_local)

        return xg_local, yg_local, zg_local

    def get_disk_3D_coordinates(self,
                                theta: Union[float, np.ndarray] = None,
                                l_over_length: Union[float, np.ndarray] = None,
                                ) -> Tuple[Union[float, np.ndarray]]:
        ### Set defaults
        if theta is None:
            theta = np.linspace(
                0,
                2 * np.pi,
                60 + 1
            )[:-1]
        if l_over_length is None:
            if self.length == 0:
                l_over_length = 0
            else:
                l_over_length = np.linspace(
                    0,
                    1,
                    4
                ).reshape((1, -1))

                theta = np.array(theta).reshape((-1, 1))

        st = np.sin(np.mod(theta, 2 * np.pi))
        ct = np.cos(np.mod(theta, 2 * np.pi))

        x = l_over_length * self.length
        y = ct * self.radius
        z = st * self.radius

        xg_local, yg_local, zg_local = self.compute_frame()

        return (
            self.xyz_c[0] + x * xg_local[0] + y * yg_local[0] + z * zg_local[0],
            self.xyz_c[1] + x * xg_local[1] + y * yg_local[1] + z * zg_local[1],
            self.xyz_c[2] + x * xg_local[2] + y * yg_local[2] + z * zg_local[2],
        )

    def translate(self,
                  xyz: Union[np.ndarray, List[float]],
                  ) -> 'Propulsor':
        """
        Returns a copy of this propulsor that has been translated by `xyz`.

        Args:
            xyz: The amount to translate the propulsor, in meters. Given in aircraft geometry axes, as with everything else.

        Returns: A copy of this propulsor, translated by `xyz`.
        """
        new_propulsor = copy.deepcopy(self)
        new_propulsor.xyz_c = new_propulsor.xyz_c + np.array(xyz)
        return new_propulsor


if __name__ == '__main__':
    p_disk = Propulsor(radius=3)
    p_can = Propulsor(length=1)



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest


def test_op_indexing():

    op = asb.OperatingPoint(
        atmosphere=asb.Atmosphere(altitude=np.linspace(0, 100)),
        velocity=np.linspace(30, 70),
        alpha=np.linspace(-5, 5),
        beta=0,
    )
    o0 = op[0]
    o1 = op[1]
    on1 = op[-1]
    on2 = op[-2]
    oslice = op[2:5]

    assert o0.atmosphere == op.atmosphere[0]
    assert o0.alpha == op.alpha[0]
    assert o0.velocity == op.velocity[0]

    assert o1.atmosphere == op.atmosphere[1]
    assert o1.alpha == op.alpha[1]
    assert o1.velocity == op.velocity[1]

    assert on1.atmosphere == op.atmosphere[-1]
    assert on1.alpha == op.alpha[-1]
    assert on1.velocity == op.velocity[-1]

    assert on2.atmosphere == op.atmosphere[-2]
    assert on2.alpha == op.alpha[-2]
    assert on2.velocity == op.velocity[-2]

    assert oslice.atmosphere == op.atmosphere[2:5]
    assert np.all(oslice.alpha == op.alpha[2:5])
    assert np.all(oslice.velocity == op.velocity[2:5])

    # # Test indexing of a simple Dynamics object
    # dyn = asb.DynamicsPointMass1DHorizontal(
    #     mass_props=asb.MassProperties(
    #         mass=1
    #     ),
    #     x_e=np.arange(10) ** 2,
    #     u_e=2 * np.arange(10),
    # )
    #
    # d0 = dyn[0]
    # d1 = dyn[1]
    # dn1 = dyn[-1]
    # dn2 = dyn[-2]
    # dslice = dyn[2:5]
    #
    # assert d0.x_e == 0
    # assert d0.u_e == 0
    # assert d1.x_e == 1
    # assert d1.u_e == 2
    # assert dn1.x_e == 81
    # assert dn1.u_e == 18
    # assert dn2.x_e == 64
    # assert dn2.u_e == 16
    # assert all(dslice.x_e == [4, 9, 16])
    # assert all(dslice.u_e == [4, 6, 8])


if __name__ == '__main__':
    test_op_indexing()



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest
from scipy import integrate

u_e_0 = 100
v_e_0 = 0
w_e_0 = -100
speed_0 = (u_e_0 ** 2 + w_e_0 ** 2) ** 0.5
gamma_0 = np.arctan2(-w_e_0, u_e_0)
track_0 = 0

time = np.linspace(0, 10, 501)


def get_trajectory(
        parameterization: type = asb.DynamicsPointMass3DCartesian,
        gravity=True,
        drag=True,
        sideforce=True,
        plot=False
):
    if parameterization is asb.DynamicsPointMass3DCartesian:
        dyn = parameterization(
            mass_props=asb.MassProperties(mass=1),
            x_e=0,
            y_e=0,
            z_e=0,
            u_e=u_e_0,
            v_e=v_e_0,
            w_e=w_e_0,
        )
    elif parameterization is asb.DynamicsPointMass3DSpeedGammaTrack:
        dyn = parameterization(
            mass_props=asb.MassProperties(mass=1),
            x_e=0,
            y_e=0,
            z_e=0,
            speed=speed_0,
            gamma=gamma_0,
            track=track_0,
        )
    else:
        raise ValueError("Bad value of `parameterization`!")

    def derivatives(t, y):
        this_dyn = dyn.get_new_instance_with_state(y)
        if gravity:
            this_dyn.add_gravity_force()
        q = 0.5 * 1.225 * this_dyn.speed ** 2
        if drag:
            this_dyn.add_force(
                Fx=-1 * (0.1) ** 2 * q,
                axes="wind"
            )
        if sideforce:
            strouhal = 0.2
            frequency = 5  # strouhal * this_dyn.speed / 0.1
            this_dyn.add_force(
                Fy=q * 1 * (0.1) ** 2 * np.sin(2 * np.pi * frequency * t),
                axes="wind"
            )

        return this_dyn.unpack_state(this_dyn.state_derivatives())

    res = integrate.solve_ivp(
        fun=derivatives,
        t_span=(time[0], time[-1]),
        t_eval=time,
        y0=dyn.unpack_state(),
        method="LSODA",
        # vectorized=True,
        rtol=1e-9,
        atol=1e-9,
    )

    dyn = dyn.get_new_instance_with_state(res.y)

    if plot:
        import matplotlib.pyplot as plt
        import aerosandbox.tools.pretty_plots as p

        fig, ax = plt.subplots()
        p.plot_color_by_value(dyn.x_e, dyn.altitude, c=dyn.speed, colorbar=True)
        p.equal()
        p.show_plot("Trajectory", "$x_e$", "$z_e$")

    return dyn


def test_final_position_Cartesian_with_sideforce():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
        drag=True,
        sideforce=True
    )
    assert dyn[-1].x_e == pytest.approx(277.3463197415092, abs=1e-2)
    assert dyn[-1].y_e == pytest.approx(10.791223276048788, abs=1e-2)
    assert dyn[-1].z_e == pytest.approx(3.1722726974414805, abs=1e-2)


def test_final_position_Cartesian_with_drag():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
        drag=True,
        sideforce=False
    )
    assert dyn[-1].x_e == pytest.approx(277.5774436945314, abs=1e-2)
    assert dyn[-1].z_e == pytest.approx(3.1722727033631886, abs=1e-2)


def test_final_position_Cartesian_no_drag():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
        drag=False,
        sideforce=False
    )
    assert dyn[-1].x_e == pytest.approx(1000, abs=1e-2)
    assert dyn[-1].z_e == pytest.approx(-509.5, abs=1e-2)


def test_final_position_SpeedGammaTrack_with_sideforce():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
        drag=True,
        sideforce=True
    )
    assert dyn[-1].x_e == pytest.approx(277.3463197415092, abs=1e-2)
    assert dyn[-1].y_e == pytest.approx(10.791223276048788, abs=1e-2)
    assert dyn[-1].z_e == pytest.approx(3.1722726974414805, abs=1e-2)


def test_final_position_SpeedGammaTrack_with_drag():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
        drag=True,
        sideforce=False
    )
    assert dyn[-1].x_e == pytest.approx(277.5774436945314, abs=1e-2)
    assert dyn[-1].z_e == pytest.approx(3.1722727033631886, abs=1e-2)


def test_final_position_SpeedGammaTrack_no_drag():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
        drag=False,
        sideforce=False
    )
    assert dyn[-1].x_e == pytest.approx(1000, abs=1e-2)
    assert dyn[-1].z_e == pytest.approx(-509.5, abs=1e-2)


def test_cross_compare_with_drag():
    dyn1 = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
        drag=True
    )
    dyn2 = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
        drag=True
    )
    assert dyn1[-1].x_e == pytest.approx(dyn2[-1].x_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].y_e == pytest.approx(dyn2[-1].y_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].z_e == pytest.approx(dyn2[-1].z_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].u_e == pytest.approx(dyn2[-1].u_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].v_e == pytest.approx(dyn2[-1].v_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].w_e == pytest.approx(dyn2[-1].w_e, abs=1e-6, rel=1e-6)


def test_cross_compare_no_drag():
    dyn1 = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
        drag=False
    )
    dyn2 = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
        drag=False
    )
    assert dyn1[-1].x_e == pytest.approx(dyn2[-1].x_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].y_e == pytest.approx(dyn2[-1].y_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].z_e == pytest.approx(dyn2[-1].z_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].u_e == pytest.approx(dyn2[-1].u_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].v_e == pytest.approx(dyn2[-1].v_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].w_e == pytest.approx(dyn2[-1].w_e, abs=1e-6, rel=1e-6)


#
if __name__ == '__main__':
    pytest.main()
    dyn = get_trajectory(
        asb.DynamicsPointMass3DSpeedGammaTrack,
        plot=True
    )



import aerosandbox.numpy as np


def critical_mach(fineness_ratio_nose: float) -> float:
    """
    Returns the transonic critical Mach number for a streamlined fuselage.

    Fitted to data from Raymer "Aircraft Design: A Conceptual Approach" 2nd Ed., Fig. 12.28.
    See figure + study + fit in: /studies/FuselageCriticalMach/

    Args:

        fineness_ratio_nose: The fineness ratio of the nose section of the fuselage.

            Specifically, fineness_ratio_nose = 2 * L_n / d, where:

                * L_n is the length from the nose to the longitudinal location at which the fuselage cross section
                becomes essentially constant, and:

                * d is the body diameter at that location.

    Returns: The critical Mach number

    """
    p = {
        'a': 11.087202397070559,
        'b': 13.469755774708842,
        'c': 4.034476257077558
    }

    mach_dd = 1 - (p["a"] / (2 * fineness_ratio_nose + p["b"])) ** p["c"]

    ### The following approximate relation is derived in W.H. Mason, "Configuration Aerodynamics", Chapter 7. Transonic Aerodynamics of Airfoils and Wings.
    ### Equation 7-8 on Page 7-19.
    ### This is in turn based on Lock's proposed empirically-derived shape of the drag rise, from Hilton, W.F., High Speed Aerodynamics, Longmans, Green & Co., London, 1952, pp. 47-49
    mach_crit = mach_dd - (0.1 / 80) ** (1 / 3)

    return mach_crit


def jorgensen_eta(fineness_ratio: float) -> float:
    """
    A fit for the eta parameter (crossflow lift multiplier) of a fuselage, as described in:

    Jorgensen, Leland Howard. "Prediction of Static Aerodynamic Characteristics for Slender Bodies
    Alone and with Lifting Surfaces to Very High Angles of Attack". NASA TR R-474. 1977.

    Fits performed in /studies/FuselageJorgensenEtaFitting/

    Args:
        fineness_ratio: The fineness ratio of the fuselage. (length / diameter)

    Returns: An estimate of eta.

    """
    x = fineness_ratio
    p = {
        '1scl': 23.009059965179222,
        '1cen': -122.76900250914575,
        '2scl': 13.006453125841258,
        '2cen': -24.367562906887436
    }
    return 1 - p["1scl"] / (x - p["1cen"]) - (p["2scl"] / (x - p["2cen"])) ** 2


def fuselage_base_drag_coefficient(mach: float) -> float:
    """
    A fit for the fuselage base drag coefficient of a cylindrical fuselage, as described in:

    MIL-HDBK-762: DESIGN OF AERODYNAMICALLY STABILIZED FREE ROCKETS:
        * Section 5-5.3.1 Body-of-Revolution Base Drag, Rocket Jet Plume-Off
        * Figure 5-140: Effects of Mach Number and Reynolds Number on Base Pressure

    Fits in /studies/FuselageBaseDragCoefficient

    Args:
        mach: Mach number [-]

    Returns: Fuselage base drag coefficient

    """

    m = mach
    p = {'a'         : 0.18024110740341143,
         'center_sup': -0.21737019935624047,
         'm_trans'   : 0.9985447737532848,
         'pc_sub'    : 0.15922582283573747,
         'pc_sup'    : 0.04698820458826384,
         'scale_sup' : 0.34978926411193456,
         'trans_str' : 9.999987483414937}

    return np.blend(
        p["trans_str"] * (m - p["m_trans"]),
        p["pc_sup"] + p["a"] * np.exp(-(p["scale_sup"] * (m - p["center_sup"])) ** 2),
        p["pc_sub"]
    )


def fuselage_form_factor(
        fineness_ratio: float,
        ratio_of_corner_radius_to_body_width: float = 0.5
):
    """
    Computes the form factor of a fuselage as a function of various geometrical parameters.

    Assumes the body cross section is a rounded square with constant-radius-of-curvature fillets.
    Body cross section can therefore vary from a true square to a true circle.

    Uses the methodology described in:

    Götten, Falk; Havermann, Marc; Braun, Carsten; Marino, Matthew; Bil, Cees.
    "Improved Form Factor for Drag Estimation of Fuselages with Various Cross Sections.
    AIAA Journal of Aircraft, 2021. DOI: 10.2514/1.C036032

    https://arc.aiaa.org/doi/10.2514/1.C036032

    Assumes fully turbulent flow. Coefficient of determination found in the paper above was 0.95.

    Note: the value returned does not account for any base separation (other than minor aft-closure separation). The
    equations were also fit to relatively-shape-optimized fuselages, and will be overly-optimistic for unoptimized
    shapes.

    Args:

        fineness_ratio: The fineness ratio of the body (length / diameter).

        ratio_of_corner_radius_to_body_width: A parameter that describes the cross-sectional shape of the fuselage.
        Precisely, this is ratio of corner radius to body width.

            * A value of 0 corresponds to a true square.

            * A value of 0.5 (default) corresponds to a true circle.

    Returns: The form factor of the body, defined as:

        C_D = C_f * form_factor * (S_wet / S_ref)

    """
    fr = fineness_ratio
    r = 2 * ratio_of_corner_radius_to_body_width

    cs1 = -0.825885 * r ** 0.411795 + 4.0001
    cs2 = -0.340977 * r ** 7.54327 - 2.27920
    cs3 = -0.013846 * r ** 1.34253 + 1.11029

    form_factor = cs1 * fr ** cs2 + cs3

    return form_factor



from aerosandbox.numpy.array import *
import pytest
import aerosandbox.numpy as np
import casadi as cas


def test_array_numpy_equivalency_1D():
    inputs = [
        1,
        2
    ]

    a = array(inputs)
    a_np = np.array(inputs)

    assert np.all(a == a_np)


def test_array_numpy_equivalency_2D():
    inputs = [
        [1, 2],
        [3, 4]
    ]

    a = array(inputs)
    a_np = np.array(inputs)

    assert np.all(a == a_np)


def test_array_casadi_1D_shape():
    a = array([cas.DM(1), cas.DM(2)])
    assert length(a) == 2


def test_can_convert_DM_to_ndarray():
    c = cas.DM([1, 2, 3])
    n = np.array(c)

    assert np.all(n == np.array([1, 2, 3]))


def test_length():
    assert length(5) == 1
    assert length(5.) == 1
    assert length([1, 2, 3]) == 3

    assert length(np.array(5)) == 1
    assert length(np.array([5])) == 1
    assert length(np.array([1, 2, 3])) == 3
    assert length(np.ones((3, 2))) == 3

    assert length(cas.MX(np.ones(5))) == 5


def test_concatenate():
    n = np.arange(10)
    c = cas.DM(n)

    assert concatenate((n, n)).shape == (20,)
    assert concatenate((n, c)).shape == (20, 1)
    assert concatenate((c, n)).shape == (20, 1)
    assert concatenate((c, c)).shape == (20, 1)

    assert concatenate((n, n, n)).shape == (30,)
    assert concatenate((c, c, c)).shape == (30, 1)


def test_stack():
    n = np.arange(10)
    c = cas.DM(n)

    assert stack((n, n)).shape == (2, 10)
    assert stack((n, n), axis=-1).shape == (10, 2)
    assert stack((n, c)).shape == (2, 10)
    assert stack((n, c), axis=-1).shape == (10, 2)
    assert stack((c, c)).shape == (2, 10)
    assert stack((c, c), axis=-1).shape == (10, 2)

    with pytest.raises(Exception):
        assert stack((n, n), axis=2)

    with pytest.raises(Exception):
        stack((c, c), axis=2)


def test_diag_onp():
    # Test on 1D array
    a = np.array([1, 2, 3])
    assert np.all(np.diag(a) == np.diag(a, k=0))
    assert np.all(
        np.diag(a, k=1) == np.array([
            [0, 1, 0, 0],
            [0, 0, 2, 0],
            [0, 0, 0, 3],
            [0, 0, 0, 0]
        ])
    )
    assert np.all(
        np.diag(a, k=-1) == np.array([
            [0, 0, 0, 0],
            [1, 0, 0, 0],
            [0, 2, 0, 0],
            [0, 0, 3, 0]
        ])
    )

    # Test on 2D square array
    b = np.array([[1, 2], [3, 4]])
    assert np.all(np.diag(b) == np.array([1, 4]))
    assert np.all(np.diag(b, k=1) == np.array([2]))
    assert np.all(np.diag(b, k=-1) == np.array([3]))

    # Test on non-square 2D array
    c = np.array([[1, 2, 3], [4, 5, 6]])
    assert np.all(np.diag(c) == np.array([1, 5]))
    assert np.all(np.diag(c, k=1) == np.array([2, 6]))
    assert np.all(np.diag(c, k=-1) == np.array([4]))


def test_diag_casadi():
    # Test on 1D array
    a = cas.SX(np.array([1, 2, 3]))
    assert np.all(np.diag(a) == np.diag(a, k=0))
    assert np.all(
        np.diag(a, k=1) == np.array([
            [0, 1, 0, 0],
            [0, 0, 2, 0],
            [0, 0, 0, 3],
            [0, 0, 0, 0]
        ])
    )
    assert np.all(
        np.diag(a, k=-1) == np.array([
            [0, 0, 0, 0],
            [1, 0, 0, 0],
            [0, 2, 0, 0],
            [0, 0, 3, 0]
        ])
    )

    # Test on 2D square array
    b = cas.SX(np.array([[1, 2], [3, 4]]))
    assert np.all(np.diag(b) == np.array([1, 4]))
    assert np.all(np.diag(b, k=1) == np.array([2]))
    assert np.all(np.diag(b, k=-1) == np.array([3]))

    # # Test on non-square 2D array
    # c = cas.SX(np.array([[1, 2, 3], [4, 5, 6]]))
    # assert np.all(np.diag(c) == np.array([1, 5]))
    # assert np.all(np.diag(c, k=1) == np.array([2, 6]))
    # assert np.all(np.diag(c, k=-1) == np.array([4]))

def test_roll_onp():
    # Test on 1D array
    a = np.arange(1, 101)  # Large array
    b = np.concatenate([np.arange(91, 101), np.arange(1, 91)])

    assert np.all(np.roll(a, 10) == b)

    # Test negative shift on 1D array
    assert np.all(np.roll(a, -10) == np.roll(a, 90))

    # Test shift larger than array size
    assert np.all(np.roll(a, 110) == np.roll(a, 10))


def test_roll_casadi():
    # Test on 1D array
    a_np = np.arange(1, 101)
    a = cas.SX(a_np)
    b = cas.SX(np.concatenate([np.arange(91, 101), np.arange(1, 91)]))

    assert np.all(cas.DM(np.roll(a, 10)) == b)

    # Test negative shift on 1D array
    assert np.all(cas.DM(np.roll(a, -10)) == cas.DM(np.roll(a, 90)))

    # Test shift larger than array size
    assert np.all(cas.DM(np.roll(a, 110)) == cas.DM(np.roll(a, 10)))


def test_roll_casadi_2d():
    # Test on 2D array
    a_np = np.reshape(np.arange(1, 101), (10, 10))  # 2D array
    a = cas.SX(a_np)

    # Shift along axis 1
    assert np.all(cas.DM(np.roll(a, 2, axis=1)) == np.roll(a_np, 2, axis=1))

    # Shift along axis 0
    assert np.all(cas.DM(np.roll(a, 2, axis=0)) == np.roll(a_np, 2, axis=0))

    # Shift along both axes
    assert np.all(cas.DM(np.roll(a, (2, 3), axis=(0, 1))) == np.roll(a_np, (2, 3), axis=(0, 1)))

    # Test on non-square 2D array
    a_np = np.reshape(np.arange(1, 201), (10, 20))  # non-square 2D array
    a = cas.SX(a_np)

    # Shift along axis 1
    assert np.all(cas.DM(np.roll(a, 2, axis=1)) == np.roll(a_np, 2, axis=1))

    # Shift along axis 0
    assert np.all(cas.DM(np.roll(a, 2, axis=0)) == np.roll(a_np, 2, axis=0))

    # Shift along both axes
    assert np.all(cas.DM(np.roll(a, (2, 3), axis=(0, 1))) == np.roll(a_np, (2, 3), axis=(0, 1)))


def test_max():
    a = cas.SX([1, 2, 3])
    b = [1, 2, 3]

    assert int(np.max(a)) == int(np.max(b))


def test_min():
    a = cas.SX([1, 2, 3])
    b = [1, 2, 3]

    assert int(np.min(a)) == int(np.min(b))


def test_reshape_1D():
    a = np.array([1, 2, 3, 4, 5, 6])
    b = cas.DM(a)

    assert b.shape == (len(a), 1)

    test_inputs = [
        -1,
        (3, 2),
        (2, 3),
        (6, 1),
        (1, 6),
        (-1),
        (6, -1),
        (-1, 6),
    ]

    for i in test_inputs:
        ra = np.reshape(a, i)
        rb = np.reshape(b, i)
        if len(ra.shape) == 1:
            ra = ra.reshape(-1, 1)

        assert np.all(ra == rb)
        assert ra.shape == rb.shape


def test_reshape_2D_vec_tall():
    a = np.array([1, 2, 3, 4, 5, 6]).reshape((6, 1))
    b = cas.DM(a)

    assert b.shape == (len(a), 1)

    test_inputs = [
        -1,
        (3, 2),
        (2, 3),
        (6, 1),
        (1, 6),
        (-1),
        (6, -1),
        (-1, 6),
    ]

    for i in test_inputs:
        ra = np.reshape(a, i)
        rb = np.reshape(b, i)
        if len(ra.shape) == 1:
            ra = ra.reshape(-1, 1)

        assert np.all(ra == rb)
        assert ra.shape == rb.shape


def test_reshape_2D_vec_wide():
    a = np.array([1, 2, 3, 4, 5, 6]).reshape((1, 6))
    b = cas.DM(a)

    assert b.shape == (1, 6)

    test_inputs = [
        -1,
        (3, 2),
        (2, 3),
        (6, 1),
        (1, 6),
        (-1),
        (6, -1),
        (-1, 6),
    ]

    for i in test_inputs:
        ra = np.reshape(a, i)
        rb = np.reshape(b, i)
        if len(ra.shape) == 1:
            ra = ra.reshape(-1, 1)

        assert np.all(ra == rb)
        assert ra.shape == rb.shape


def test_reshape_2D():
    a_np = np.array([
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
        [10, 11, 12],
    ])
    a_cas = cas.DM(a_np)

    test_inputs = [
        -1,
        (4, 3),
        (3, 4),
        (12, 1),
        (1, 12),
        (-1),
        (12, -1),
        (-1, 12),
    ]

    for i in test_inputs:
        res_np = np.reshape(a_np, i)
        res_cas = np.reshape(a_cas, i)
        if len(res_np.shape) == 1:
            res_np = res_np.reshape(-1, 1)

        assert np.all(res_np == res_cas)
        assert res_np.shape == res_cas.shape

    for i in test_inputs:
        res_np = np.reshape(a_np, i, order="F")
        res_cas = np.reshape(a_cas, i, order="F")
        if len(res_np.shape) == 1:
            res_np = res_np.reshape(-1, 1)

        assert np.all(res_np == res_cas)
        assert res_np.shape == res_cas.shape


def test_assert_equal_shape():
    a = np.array([1, 2, 3])
    b = cas.DM(a)

    np.assert_equal_shape([
        a,
        a
    ])
    np.assert_equal_shape({
        "thing1": a,
        "thing2": a,
    })
    with pytest.raises(ValueError):
        np.assert_equal_shape([
            np.array([1, 2, 3]),
            np.array([1, 2, 3, 4])
        ])
        np.assert_equal_shape({
            "thing1": np.array([1, 2, 3]),
            "thing2": np.array([1, 2, 3, 4])
        })
    np.assert_equal_shape([
        2,
        3,
        4
    ])


if __name__ == '__main__':
    # # Test on 1D array
    # a_np = np.arange(1, 101)
    # a = cas.SX(a_np)
    # b = cas.SX(np.concatenate([np.arange(91, 101), np.arange(1, 91)]))
    #
    # s1 = np.roll(a, -10, axis=0)
    # s2 = np.roll(a, 90, axis=0)
    #
    # assert np.all(cas.DM(s1) == cas.DM(s2))

    pytest.main()



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest


def test_airplane_optimization():
    def get_aero(alpha, taper_ratio):
        airplane = asb.Airplane(
            wings=[
                asb.Wing(
                    symmetric=True,
                    xsecs=[
                        asb.WingXSec(
                            xyz_le=[-0.25, 0, 0],
                            chord=1,
                        ),
                        asb.WingXSec(
                            xyz_le=[-0.25 * taper_ratio, 1, 0],
                            chord=taper_ratio,
                        )
                    ]
                )
            ]
        )
        op_point = asb.OperatingPoint(
            velocity=1,
            alpha=alpha,
        )

        vlm = asb.VortexLatticeMethod(
            airplane,
            op_point,
            chordwise_resolution=6,
            spanwise_resolution=6,
        )
        return vlm.run()

    # tr = np.linspace(0.01, 1)
    # aeros = np.vectorize(get_aero)(tr)
    # import matplotlib.pyplot as plt
    # import aerosandbox.tools.pretty_plots as p
    # fig, ax = plt.subplots()
    # plt.plot(tr, [a['CL'] / a['CD'] for a in aeros])
    # p.show_plot()

    opti = asb.Opti()
    alpha = opti.variable(0, lower_bound=-90, upper_bound=90)
    taper_ratio = opti.variable(0.7, lower_bound=0.0001, upper_bound=1)
    aero = get_aero(alpha, taper_ratio)
    CD0 = 0.01
    LD = aero["CL"] / (aero["CD"] + CD0)
    opti.minimize(-LD)
    sol = opti.solve()

    assert sol(alpha) == pytest.approx(5.5, abs=1)


if __name__ == '__main__':

    pytest.main()



from aerosandbox.geometry.airplane import Airplane
from aerosandbox.performance import OperatingPoint
from aerosandbox.weights import MassProperties
import aerosandbox.numpy as np


def get_modes(
        airplane: Airplane,
        op_point: OperatingPoint,
        mass_props: MassProperties,
        aero,
        g=9.81,
):
    Q = op_point.dynamic_pressure()
    S = airplane.s_ref
    c = airplane.c_ref
    b = airplane.b_ref
    QS = Q * S
    m = mass_props.mass
    Ixx = mass_props.Ixx
    Iyy = mass_props.Iyy
    Izz = mass_props.Izz
    u0 = op_point.velocity

    Cxu = -2 * aero["CD"]
    Czu = -2 * aero["CL"]

    Cmq = aero["Cmq"]
    Cma = aero["Cma"]

    X_u = QS / m / u0 * Cxu  # Units: 1/sec
    # X_w = QS / m / u0 * Cxa
    # X_q = QS / m * c / (2 * u0) * Cxq
    Z_u = QS / m / u0 * Czu  # Units: 1/sec
    # Z_w = QS / m / u0 * Cza
    # Z_q = QS / m * c / (2 * u0) * Czq
    # M_u = QS * c / Iyy / u0 * Cmu
    M_w = QS * c / Iyy / u0 * Cma  # Units: 1/(meter * sec)
    M_q = QS * c / Iyy * c / (2 * u0) * Cmq  # Units: 1/sec

    modes = {}

    def get_mode_info(
            sigma,
            omega_squared,
    ):
        is_oscillatory = omega_squared > 0
        return {
            "eigenvalue_real": sigma + np.where(
                is_oscillatory,
                0,
                np.abs(omega_squared + 1e-100) ** 0.5,
            ),
            "eigenvalue_imag": np.where(
                is_oscillatory,
                np.abs(omega_squared + 1e-100) ** 0.5,
                0,
            ),
        }

    ##### Longitudinal modes

    ### Phugoid
    modes['phugoid'] = {  # FVA, Eq. 9.55-9.57
        **get_mode_info(
            sigma=X_u / 2,
            omega_squared=-(X_u ** 2) / 4 - g * Z_u / u0
        ),
        "eigenvalue_imag_approx": 2 ** 0.5 * g / u0,
        "damping_ratio_approx"  : 2 ** -0.5 * aero["CD"] / aero["CL"],
    }
    #
    # ### Short-period
    modes['short_period'] = get_mode_info(
        sigma=0.5 * M_q,
        omega_squared=-(M_q ** 2) / 4 - u0 * M_w
    )

    ##### Lateral modes

    ### Roll subsidence
    modes['roll_subsidence'] = {  # FVA, Eq. 9.63
        "eigenvalue_real": (
                QS * b ** 2 / (2 * Ixx * u0) * aero["Clp"]
        ),
        "eigenvalue_imag": 0,
        "damping_ratio"  : 1,
    }

    ### Dutch roll

    modes['dutch_roll'] = get_mode_info(  # FVA, Eq. 9.68
        sigma=(
                QS * b ** 2 /
                (2 * Izz * u0) *
                (aero["Cnr"] + Izz / (m * b ** 2) * aero["CYb"])
        ),
        omega_squared=(
                QS * b / Izz *
                (
                        aero["Cnb"] + (
                        op_point.atmosphere.density() * S * b / (4 * m) *
                        (aero["CYb"] * aero["Cnr"] - aero["Cnb"] * aero["CYr"])
                )
                )
        )
    )

    ### Spiral
    spiral_parameter = (aero["Cnr"] - aero["Cnb"] * aero["Clr"] / aero["Clb"])  # FVA, Eq. 9.66
    modes['spiral'] = {  # FVA, Eq. 9.66
        "eigenvalue_real": (
                QS * b ** 2 / (2 * Izz * u0) *
                spiral_parameter
        ),
        "eigenvalue_imag": 0,
    }

    ### Compute damping ratios of all modes
    for mode_name, mode_data in modes.items():
        modes[mode_name]['damping_ratio'] = (
                -mode_data['eigenvalue_real'] /
                (mode_data['eigenvalue_real'] ** 2 + mode_data['eigenvalue_imag'] ** 2) ** 0.5
        )

    return modes


if __name__ == '__main__':
    import aerosandbox as asb
    import aerosandbox.numpy as np
    from aerosandbox.tools import units as u

    from pprint import pprint

    # Numbers below are from:
    # Caughey, David A., "Introduction to Aircraft Stability and Control, Course Notes for M&AE 5070", 2011
    # https://courses.cit.cornell.edu/mae5070/Caughey_2011_04.pdf
    airplane = asb.Airplane(
        name='Boeing 737-800',
        s_ref=1260 * u.foot ** 2,
        c_ref=11 * u.foot,
        b_ref=113 * u.foot,
    )

    aero = dict(
        CL=1.83443,
        CD=0.13037,
        Cm=0,
        CLa=5.542930,
        Cma=-2.044696,
        CYb=-1.103873,
        Clb=-0.374933,
        Cnb=0.239877,
        CYp=0.800161,
        Clp=-0.449404,
        Cnp=-0.255028,
        CLq=18.973344,
        Cmq=-74.997742,
        CYr=0.796001,
        Clr=0.364638,
        Cnr=-0.434410
    )

    mass_props_TOGW = asb.MassProperties(
        mass=77146,
        x_cg=65.2686 * u.foot,
        y_cg=0.0,
        z_cg=1.16559 * u.foot,
        Ixx=706684,
        Iyy=0.270824e7,
        Izz=0.330763e7,
        Ixy=0.0,
        Iyz=0.0,
        Ixz=26994.4,
    )

    op_point = asb.OperatingPoint(
        atmosphere=asb.Atmosphere(
            altitude=2438.399975619396,
            method='differentiable',
        ),
        velocity=85.64176936131635,
    )

    assert np.allclose(
        aero["CL"],
        mass_props_TOGW.mass * 9.81 / op_point.dynamic_pressure() / airplane.s_ref,
        rtol=0.001
    )

    eigenvalues_from_AVL = {
        'phugoid'        : -0.0171382 + 0.145072j, # Real is wrong (2x)
        'short_period'   : -0.439841 + 0.842195j, # Pretty close
        'roll_subsidence': -1.35132, # get_modes says -1.81
        'dutch_roll'     : -0.385418 + 1.52695j, # Imag is wrong (1.5x)
        'spiral'         : -0.0573017, # Too small, get_modes says -0.17
    }

    pprint(
        get_modes(
            airplane=airplane,
            op_point=op_point,
            mass_props=mass_props_TOGW,
            aero=aero
        )
    )



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest


def test_roundtrip_conversion_similarity():

    for name in [
        "dae11",
        "rae2822",
        "naca0012",
    ]:
        af = asb.Airfoil(name)
        kaf = af.to_kulfan_airfoil()
        kaf_a = kaf.to_airfoil()

        similarity = af.jaccard_similarity(kaf_a)
        print(f"Airfoil: {name}, Similarity: {similarity}")

        assert similarity > 0.995


def test_TE_angle():
    kaf = asb.KulfanAirfoil("naca4412")
    af = kaf.to_airfoil()

    assert kaf.TE_angle() == pytest.approx(af.TE_angle(), rel=0.1)

def test_LE_radius():
    kaf = asb.KulfanAirfoil("naca4412")
    af = kaf.to_airfoil()

    assert kaf.LE_radius() == pytest.approx(af.LE_radius(), rel=0.1)

def test_area():
    kaf = asb.KulfanAirfoil("naca4412")
    af = kaf.to_airfoil()

    assert kaf.area() == pytest.approx(af.area(), rel=0.01)

if __name__ == '__main__':
    test_roundtrip_conversion_similarity()
    pytest.main([__file__])



import aerosandbox.numpy as np
from aerosandbox.tools.string_formatting import eng_string
import copy
from typing import Union

universal_gas_constant = 8.31432  # J/(mol*K); universal gas constant


class PerfectGas:
    """
    Provides a class for an ideal, calorically perfect gas.

    Specifically, this gas:
        * Has PV = nRT (ideal)
        * Has constant heat capacities C_V, C_P (independent of temperature and pressure).
        * Is in thermodynamic equilibrium
        * Is not chemically reacting
        * Has internal energy and enthalpy purely as functions of temperature
    """

    def __init__(self,
                 pressure: Union[float, np.ndarray] = 101325,
                 temperature: Union[float, np.ndarray] = 273.15 + 15,
                 specific_heat_constant_pressure: float = 1006,
                 specific_heat_constant_volume: float = 717,
                 molecular_mass: float = 28.9644e-3,
                 effective_collision_diameter: float = 0.365e-9,
                 ):
        """

        Args:

            pressure: Pressure of the gas, in Pascals

            temperature: Temperature of the gas, in Kelvin

            specific_heat_constant_pressure: Specific heat at constant pressure, also known as C_p. In J/kg-K.

            specific_heat_constant_volume: Specific heat at constant volume, also known as C_v. In J/kg-K.

            molecular_mass: Molecular mass of the gas, in kg/mol

            effective_collision_diameter: Effective collision diameter of a molecule, in meters.

        """
        self.pressure = pressure
        self.temperature = temperature
        self.specific_heat_constant_pressure = specific_heat_constant_pressure
        self.specific_heat_constant_volume = specific_heat_constant_volume
        self.molecular_mass = molecular_mass
        self.effective_collision_diameter = effective_collision_diameter

    def __repr__(self) -> str:
        f = lambda s, u: eng_string(s, unit=u, format="%.6g")

        return f"Gas (P = {f(self.pressure, 'Pa')}, T = {self.temperature:.6g} K, ρ = {self.density:.6g} kg/m^3, Pv^gamma = {self.pressure * self.specific_volume ** self.ratio_of_specific_heats: .6g})"

    @property
    def density(self):
        return self.pressure / (self.temperature * self.specific_gas_constant)

    @property
    def speed_of_sound(self):
        return (self.ratio_of_specific_heats * self.specific_gas_constant * self.temperature) ** 0.5

    @property
    def specific_gas_constant(self):
        return universal_gas_constant / self.molecular_mass

    @property
    def ratio_of_specific_heats(self):
        return self.specific_heat_constant_pressure / self.specific_heat_constant_volume

    def specific_enthalpy_change(self, start_temperature, end_temperature):
        """
        Returns the change in specific enthalpy that would occur from a given temperature change via a thermodynamic
        process.

        Args:
            start_temperature: Starting temperature [K]
            end_temperature: Ending temperature [K]

        Returns: The change in specific enthalpy, in J/kg.

        """
        return self.specific_heat_constant_pressure * (end_temperature - start_temperature)

    def specific_internal_energy_change(self, start_temperature, end_temperature):
        """
        Returns the change in specific internal energy that would occur from a given temperature change via a
        thermodynamic process.

        Args:
            start_temperature: Starting temperature [K]
            end_temperature: Ending temperature [K]

        Returns: The change in specific internal energy, in J/kg.

        """
        return self.specific_heat_constant_volume * (end_temperature - start_temperature)

    @property
    def specific_volume(self):
        """
        Gives the specific volume, often denoted `v`.

        (Note the lowercase; "V" is often the volume of a specific amount of gas, and this presents a potential point
        of confusion.)
        """
        return 1 / self.density

    @property
    def specific_enthalpy(self):
        """
        Gives the specific enthalpy, often denoted `h`.

        Enthalpy here is in units of J/kg.
        """
        return self.specific_enthalpy_change(start_temperature=0, end_temperature=self.temperature)

    @property
    def specific_internal_energy(self):
        """
        Gives the specific internal energy, often denoted `u`.

        Internal energy here is in units of J/kg.
        """
        return self.specific_internal_energy_change(start_temperature=0, end_temperature=self.temperature)

    def process(self,
                process: str = "isentropic",
                new_pressure: float = None,
                new_temperature: float = None,
                new_density: float = None,
                enthalpy_addition_at_constant_pressure: float = None,
                enthalpy_addition_at_constant_volume: float = None,
                polytropic_n: float = None,
                inplace=False
                ) -> "PerfectGas":
        """
        Puts this gas under a thermodynamic process.

        Equations here: https://en.wikipedia.org/wiki/Ideal_gas_law

        Args:

            process: Type of process. One of:

                * "isobaric"
                * "isochoric"
                * "isothermal"
                * "isentropic"
                * "polytropic"

                The `process` must be specified.

            You must specifiy exactly one of the following arguments:
                * `new_pressure`: the new pressure after the process [Pa].
                * `new_temperature`: the new temperature after the process [K]
                * `new_density`: the new density after the process [kg/m^3]
                * `enthalpy_addition_at_constant_pressure`: [J/kg]
                * `enthalpy_addition_at_constant_volume`: [J/kg]

            polytropic_n: If you specified the process type to be "polytropic", you must provide the polytropic index
            `n` to be used here. (Reminder: PV^n = constant)

            inplace: Specifies whether to return the result in-place or to allocate a new PerfectGas object in memory
            for the result.

        Returns:

            If `inplace` is False (default), returns a new PerfectGas object that represents the gas after the change.

            If `inplace` is True, nothing is returned.

        """

        pressure_specified = new_pressure is not None
        temperature_specified = new_temperature is not None
        density_specified = new_density is not None
        enthalpy_at_pressure_specified = enthalpy_addition_at_constant_pressure is not None
        enthalpy_at_volume_specified = enthalpy_addition_at_constant_volume is not None

        number_of_conditions_specified = (
                pressure_specified +
                temperature_specified +
                density_specified +
                enthalpy_at_pressure_specified +
                enthalpy_at_volume_specified
        )

        if number_of_conditions_specified != 1:
            raise ValueError("You must specify exactly one of the following arguments:\n" + "\n".join([
                "\t* `new_pressure`",
                "\t* `new_temperature`",
                "\t* `new_density`",
                "\t* `enthalpy_addition_at_constant_pressure`",
                "\t* `enthalpy_addition_at_constant_volume`",
            ]))

        if enthalpy_at_pressure_specified:
            new_temperature = self.temperature + enthalpy_addition_at_constant_pressure / self.specific_heat_constant_pressure
        elif enthalpy_at_volume_specified:
            new_temperature = self.temperature + enthalpy_addition_at_constant_volume / self.specific_heat_constant_volume

        if pressure_specified:
            P_ratio = new_pressure / self.pressure
        elif temperature_specified:
            T_ratio = new_temperature / self.temperature
        elif density_specified:
            V_ratio = 1 / (new_density / self.density)

        if process == "isobaric":

            new_pressure = self.pressure

            if pressure_specified:
                raise ValueError("Can't specify pressure change for an isobaric process!")

            elif density_specified:
                new_temperature = self.temperature * V_ratio

            elif temperature_specified:
                pass

        elif process == "isochoric":

            if pressure_specified:
                new_temperature = self.temperature * P_ratio

            elif density_specified:
                raise ValueError("Can't specify density change for an isochoric process!")

            elif temperature_specified:
                new_pressure = self.pressure * T_ratio

        elif process == "isothermal":

            new_temperature = self.temperature

            if pressure_specified:
                pass

            elif density_specified:
                new_pressure = self.pressure / V_ratio

            elif temperature_specified:
                raise ValueError("Can't specify temperature change for an isothermal process!")

        elif process == "isentropic":

            gam = self.ratio_of_specific_heats

            if pressure_specified:
                new_temperature = self.temperature * P_ratio ** ((gam - 1) / gam)

            elif density_specified:
                new_pressure = self.pressure * V_ratio ** -gam
                new_temperature = self.temperature * V_ratio ** (1 - gam)

            elif temperature_specified:
                new_pressure = self.pressure * T_ratio ** (gam / (gam - 1))

        elif process == "polytropic":

            if polytropic_n is None:
                raise ValueError("If the process is polytropic, then the polytropic index `n` must be specified.")

            n = polytropic_n

            if pressure_specified:
                new_temperature = self.temperature * P_ratio ** ((n - 1) / n)

            elif density_specified:
                new_pressure = self.pressure * V_ratio ** -n
                new_temperature = self.temperature * V_ratio ** (1 - n)

            elif temperature_specified:
                new_pressure = self.pressure * T_ratio ** (n / (n - 1))

        elif process == "isenthalpic":
            raise NotImplementedError()

        else:
            raise ValueError("Bad value of `process`!")

        if inplace:
            self.pressure = new_pressure
            self.temperature = new_temperature

        else:
            return PerfectGas(
                pressure=new_pressure,
                temperature=new_temperature,
                specific_heat_constant_pressure=self.specific_heat_constant_pressure,
                specific_heat_constant_volume=self.specific_heat_constant_volume,
                molecular_mass=self.molecular_mass,
                effective_collision_diameter=self.effective_collision_diameter
            )


if __name__ == '__main__':

    ### Carnot
    g = []

    g.append(PerfectGas(pressure=100e3, temperature=300))
    g.append(g[-1].process("isothermal", new_density=0.5))
    g.append(g[-1].process("isentropic", new_density=0.25))
    g.append(g[-1].process("isothermal", new_density=0.58))
    g.append(g[-1].process("isentropic", new_temperature=300))

    for i in range(len(g)):
        print(f"After Process {i}: {g[i]}")



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest

rtol = 0.10

def make_fuselage(
        diameter = 0.1,
        length = 10,
        alpha_geometric=0.,
        beta_geometric=0.,
        N=101,
):
    return asb.Fuselage(
        xsecs=[
            asb.FuselageXSec(
                xyz_c=[
                    xi * np.cosd(alpha_geometric) * np.cosd(beta_geometric),
                    xi * np.sind(beta_geometric),
                    xi * np.sind(-alpha_geometric)
                ],
                radius=diameter / 2
            )
            for xi in [0, length]
        ]
    ).subdivide_sections(N)

def test_90_deg_crossflow():
    diameter = 0.1
    length = 10

    fuselage = make_fuselage(
        diameter=diameter,
        length=length,
        alpha_geometric=0
    )

    airplane = asb.Airplane(
        fuselages=[fuselage],
        s_ref=diameter * length,
        c_ref=length,
        b_ref=diameter,
    )

    op_point = asb.OperatingPoint(
            velocity=50,
            alpha=90,
            beta=0,
        )

    aero = asb.AeroBuildup(
        airplane=airplane,
        op_point=op_point,
        xyz_ref=np.array([length / 2, 0, 0])
    ).run()

    print(aero)

    from aerosandbox.library.aerodynamics.viscous import Cd_cylinder

    CD_expected = Cd_cylinder(
        Re_D=op_point.reynolds(diameter),
        mach=op_point.mach()
    )
    print(CD_expected)

    assert aero["CD"] == pytest.approx(
        CD_expected, rel=0.1
    )


if __name__ == '__main__':
    pass
    # make_fuselage().draw_three_view()
    test_90_deg_crossflow()


import aerosandbox.numpy as np
from aerosandbox.dynamics.point_mass.common_point_mass import _DynamicsPointMassBaseClass
from abc import ABC, abstractmethod, abstractproperty
from typing import Union, Tuple
from aerosandbox import OperatingPoint, Atmosphere


class _DynamicsRigidBodyBaseClass(_DynamicsPointMassBaseClass, ABC):

    # TODO: add method for force at offset (i.e., add moment and force)

    @abstractmethod
    def add_moment(self,
                   Mx: Union[float, np.ndarray] = 0,
                   My: Union[float, np.ndarray] = 0,
                   Mz: Union[float, np.ndarray] = 0,
                   axes="body",
                   ) -> None:
        """
        Adds a moment (in whichever axis system you choose) to this Dynamics instance.

        Args:
            Mx: Moment about the x-axis in the axis system chosen. Assumed these moments are applied about the center of mass. [Nm]
            My: Moment about the y-axis in the axis system chosen. Assumed these moments are applied about the center of mass. [Nm]
            Mz: Moment about the z-axis in the axis system chosen. Assumed these moments are applied about the center of mass. [Nm]
            axes: The axis system that the specified moment is in. One of:
                * "geometry"
                * "body"
                * "wind"
                * "stability"
                * "earth"

        Returns: None (in-place)

        """
        pass

    @property
    def op_point(self):
        return OperatingPoint(
            atmosphere=Atmosphere(altitude=self.altitude),
            velocity=self.speed,
            alpha=self.alpha,
            beta=self.beta,
            p=self.p,
            q=self.q,
            r=self.r,
        )

    @property
    def alpha(self):
        """The angle of attack, in degrees."""
        return np.arctan2d(
            self.w_b,
            self.u_b
        )

    @property
    def beta(self):
        """The sideslip angle, in degrees."""
        return np.arctan2d(
            self.v_b,
            (
                    self.u_b ** 2 +
                    self.w_b ** 2
            ) ** 0.5
        )

    @property
    def rotational_kinetic_energy(self):
        return 0.5 * (
                self.mass_props.Ixx * self.p ** 2 +
                self.mass_props.Iyy * self.q ** 2 +
                self.mass_props.Izz * self.r ** 2
        )

    @property
    def kinetic_energy(self):
        return self.translational_kinetic_energy + self.rotational_kinetic_energy



import aerosandbox.numpy as np


def temperature_over_total_temperature(
        mach,
        gamma=1.4
):
    """
    Gives T/T_t, the ratio of static temperature to total temperature.

    Args:
        mach: Mach number [-]
        gamma: The ratio of specific heats. 1.4 for air across most temperature ranges of interest.
    """
    return (1 + (gamma - 1) / 2 * mach ** 2) ** -1


def pressure_over_total_pressure(
        mach,
        gamma=1.4
):
    """
    Gives P/P_t, the ratio of static pressure to total pressure.

    Args:
        mach: Mach number [-]
        gamma: The ratio of specific heats. 1.4 for air across most temperature ranges of interest.
    """
    return temperature_over_total_temperature(mach=mach, gamma=gamma) ** (gamma / (gamma - 1))


def density_over_total_density(
        mach,
        gamma=1.4
):
    """
    Gives rho/rho_t, the ratio of density to density after isentropic compression.

    Args:
        mach: Mach number [-]
        gamma: The ratio of specific heats. 1.4 for air across most temperature ranges of interest.
    """
    return temperature_over_total_temperature(mach=mach, gamma=gamma) ** (1 / (gamma - 1))


def area_over_choked_area(
        mach,
        gamma=1.4
):
    """
    Gives A/A^* (where A^* is "A-star"), the ratio of cross-sectional flow area to the cross-sectional flow area that would result in choked (M=1) flow.

    Applicable to 1D isentropic nozzle flow.

    Args:
        mach: Mach number [-]
        gamma: The ratio of specific heats. 1.4 for air across most temperature ranges of interest.
    """
    gp1 = gamma + 1
    gm1 = gamma - 1

    return (
            (gp1 / 2) ** (-gp1 / (2 * gm1)) *
            (1 + gm1 / 2 * mach ** 2) ** (gp1 / (2 * gm1)) / mach
    )


if __name__ == '__main__':

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    mach = np.linspace(0.5, 3, 500)

    fig, ax = plt.subplots()

    for name, data in {
        "$T/T_t$"       : temperature_over_total_temperature(mach),
        "$P/P_t$"       : pressure_over_total_pressure(mach),
        "$A/A^*$"       : area_over_choked_area(mach),
        r"$\rho/\rho_t$": density_over_total_density(mach),
    }.items():
        plt.plot(mach, data, label=name)
    p.show_plot()



from typing import Union, Iterable, Tuple, Optional, Callable
import matplotlib.pyplot as plt

import numpy as np
from aerosandbox.tools.statistics import time_series_uncertainty_quantification as tsuq


def plot_with_bootstrapped_uncertainty(
        x: np.ndarray,
        y: np.ndarray,
        ci: Optional[Union[float, Iterable[float], np.ndarray]] = 0.95,
        x_stdev: Union[None, float] = 0.,
        y_stdev: Union[None, float] = None,
        color: Optional[Union[str, Tuple[float]]] = None,
        draw_data: bool = True,
        label_line: Union[bool, str] = "Best Estimate",
        label_ci: bool = True,
        label_data: Union[bool, str] = "Raw Data",
        line_alpha: float = 0.9,
        ci_to_alpha_mapping: Callable[[float], float] = lambda ci: 0.8 * (1 - ci) ** 0.4,
        n_bootstraps=2000,
        n_fit_points=500,
        spline_degree=3,
        normalize: bool=True,
        x_log_scale: bool = False,
        y_log_scale: bool = False,
):
    x = np.array(x)
    y = np.array(y)

    ### Log-transform the data if desired
    if x_log_scale:
        x = np.log(x)
    if y_log_scale:
        y = np.log(y)

    ### Make sure `ci` is a NumPy array
    if ci is None:
        ci = []
    else:
        try:
            iter(ci)
        except TypeError:
            ci = [ci]
    ci = np.array(ci)

    ### Make sure `ci` is sorted
    ci = np.sort(ci)

    ### Make sure `ci` is in bounds
    if not (np.all(ci > 0) and np.all(ci < 1)):
        raise ValueError("Confidence interval values in `ci` should all be in the range of (0, 1).")

    ### Do the bootstrap fits
    x_fit, y_bootstrap_fits = tsuq.bootstrap_fits(
        x=x,
        y=y,
        x_noise_stdev=x_stdev,
        y_noise_stdev=y_stdev,
        n_bootstraps=n_bootstraps,
        fit_points=n_fit_points,
        spline_degree=spline_degree,
        normalize=normalize,
    )

    ### Undo the log-transform if desired
    if x_log_scale:
        x = np.exp(x)
        x_fit = np.exp(x_fit)
    if y_log_scale:
        y = np.exp(y)
        y_bootstrap_fits = np.exp(y_bootstrap_fits)

    ### Plot the best-estimator line
    line, = plt.plot(
        x_fit,
        np.nanquantile(y_bootstrap_fits, q=0.5, axis=0),
        color=color,
        label=label_line,
        zorder=2,
        alpha=line_alpha,
    )
    if color is None:
        color = line.get_color()

    if x_log_scale:
        plt.xscale('log')
    if y_log_scale:
        plt.yscale('log')

    ### Plot the confidence intervals
    if len(ci) != 0:

        ### Using the method of equal-tails confidence intervals
        lower_quantiles = np.concatenate([[0.5], (1 - ci) / 2])
        upper_quantiles = np.concatenate([[0.5], 1 - (1 - ci) / 2])

        lower_ci = np.nanquantile(y_bootstrap_fits, q=lower_quantiles, axis=0)
        upper_ci = np.nanquantile(y_bootstrap_fits, q=upper_quantiles, axis=0)

        for i, ci_val in enumerate(ci):
            settings = dict(
                color=color,
                alpha=ci_to_alpha_mapping(ci_val),
                linewidth=0,
                zorder=1.5
            )
            plt.fill_between(
                x_fit,
                lower_ci[i],
                lower_ci[i + 1],
                label=f"{ci_val:.0%} CI" if label_ci else None,
                **settings
            )
            plt.fill_between(
                x_fit,
                upper_ci[i],
                upper_ci[i + 1],
                **settings
            )

    ### Plot the data
    if draw_data:
        plt.plot(
            x,
            y,
            ".k",
            label=label_data,
            alpha=0.25,
            markersize=5,
            markeredgewidth=0,
            zorder=1,
        )
    return x_fit, y_bootstrap_fits


if __name__ == '__main__':

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    np.random.seed(0)

    ### Generate data
    x = np.linspace(0, 20, 1001)
    y_true = np.sin(x - 5)  # np.sin(x)

    y_stdev = 0.5

    y_noisy = y_true + y_stdev * np.random.randn(len(x))

    ### Plot spline regression
    fig, ax = plt.subplots(dpi=300)
    x_fit, y_bootstrap_fits = plot_with_bootstrapped_uncertainty(
        x,
        y_noisy,
        ci=[0.75, 0.95],
        label_line="Best Estimate",
        label_data="Data (True Function + Noise)",
    )
    ax.plot(x, y_true, "k--", label="True Function (Hidden)", alpha=0.8, zorder=1)
    plt.legend(ncols=2)

    p.show_plot(
        "Spline Bootstrapping Test",
        r"$x$",
        r"$y$",
        legend=False
    )

    ### Generate data
    x = np.geomspace(10, 1000, 1000)
    y_true = 3 * x ** 0.5

    y_stdev = 0.1

    y_noisy = y_true * y_stdev * np.random.lognormal(size=len(x))

    fig, ax = plt.subplots()
    x_fit, y_bootstrap_fits = plot_with_bootstrapped_uncertainty(
        x,
        y_noisy,
        ci=[0.75, 0.95],
        label_line="Best Estimate",
        label_data="Data (True Function + Noise)",
        # normalize=False,
        x_log_scale=True,
        y_log_scale=True,
    )
    p.show_plot()
